"""Generate XML Schema Definition (XSD) corresponding to the meta-model."""

import re
import xml.etree.ElementTree as ET

# noinspection PyUnresolvedReferences
import xml.dom.minidom
from typing import TextIO, MutableMapping, Optional, Tuple, List, Sequence, Any

from icontract import ensure, require
import greenery

import aas_core_codegen.xsd
from aas_core_codegen import (
    naming,
    specific_implementations,
    intermediate,
    run,
    infer_for_schema,
)
from aas_core_codegen.common import Error, assert_never, Identifier
from aas_core_codegen.xsd import naming as xsd_naming
from aas_core_codegen.parse import retree as parse_retree

assert aas_core_codegen.xsd.__doc__ == __doc__


def _define_for_enumeration(enumeration: intermediate.Enumeration) -> List[ET.Element]:
    """
    Generate the definitions for an ``enumeration``.
    The root element is to be *extended* with the resulting list.
    """
    restriction = ET.Element("xs:restriction", {"base": "xs:string"})
    for literal in enumeration.literals:
        restriction.append(ET.Element("xs:enumeration", {"value": literal.value}))

    element = ET.Element(
        "xs:simpleType", {"name": xsd_naming.type_name(enumeration.name)}
    )
    element.append(restriction)

    return [element]


_PRIMITIVE_MAP = {
    intermediate.PrimitiveType.BOOL: "xs:boolean",
    intermediate.PrimitiveType.INT: "xs:long",
    intermediate.PrimitiveType.FLOAT: "xs:double",
    intermediate.PrimitiveType.STR: "xs:string",
    intermediate.PrimitiveType.BYTEARRAY: "xs:base64Binary",
}
assert all(literal in _PRIMITIVE_MAP for literal in intermediate.PrimitiveType)

# noinspection RegExpSimplifiable
_ESCAPE_BACKSLASH_X_RE = re.compile(r"\\x([a-fA-f0-9]{2})")


def _undo_escaping_backslash_x_in_pattern(pattern: str) -> str:
    """
    Undo the escaping of `\\x??` in the ``pattern``.

    This is necessary since XML Schema Validators do not know how to handle such escape
    sequences in the patterns and need the verbatim characters.
    """
    parts = []  # type: List[str]
    cursor = None  # type: Optional[int]
    for mtch in re.finditer(_ESCAPE_BACKSLASH_X_RE, pattern):
        if cursor is None:
            parts.append(pattern[: mtch.start()])
        else:
            parts.append(pattern[cursor : mtch.start()])

        ascii_code = int(mtch.group(1), base=16)
        character = chr(ascii_code)
        parts.append(character)
        cursor = mtch.end()

    if cursor is None:
        parts.append(pattern)
    else:
        if cursor < len(pattern):
            parts.append(pattern[cursor:])

    return "".join(parts)


# noinspection RegExpSimplifiable
_ESCAPE_BACKSLASH_X_U_U_RE = re.compile(
    r"(\\x([a-fA-f0-9]{2})|\\u([a-fA-f0-9]{4})|\\U([a-fA-f0-9]{8}))"
)


def _undo_escaping_backslash_x_u_and_U_in_pattern(pattern: str) -> str:
    """
    Undo the escaping of ``\\x??``, ``\\u????`` and ``\\U????????`` in the ``pattern``.

    This is necessary since Greenery does not know how to handle such escape
    sequences in the patterns and need the verbatim characters.
    """
    parts = []  # type: List[str]
    cursor = None  # type: Optional[int]
    for mtch in re.finditer(_ESCAPE_BACKSLASH_X_U_U_RE, pattern):
        if cursor is None:
            parts.append(pattern[: mtch.start()])
        else:
            parts.append(pattern[cursor : mtch.start()])

        substring = mtch.group(0)
        assert len(substring) > 2
        assert substring[0] == "\\"

        hex_code = substring[2:]
        code_point = int(hex_code, base=16)
        character = chr(code_point)
        parts.append(character)
        cursor = mtch.end()

    if cursor is None:
        parts.append(pattern)
    else:
        if cursor < len(pattern):
            parts.append(pattern[cursor:])

    return "".join(parts)


class _AnchorRemover(parse_retree.PassThroughVisitor):
    """
    Remove anchors from a regex in-place.

    We need to remove the anchors (``^``, ``$``) since patterns in the XSD are always
    anchored.

    This is necessary since otherwise the schema validation fails.
    See: https://stackoverflow.com/questions/4367914/regular-expression-in-xml-schema-definition-fails
    """

    def visit_concatenation(self, node: parse_retree.Concatenation) -> None:
        """Visit the ``concatenation``."""
        new_concatenants = []  # type: List[parse_retree.Term]
        for concatenant in node.concatenants:
            if not (
                isinstance(concatenant.value, parse_retree.Symbol)
                and concatenant.value.kind
                in (parse_retree.SymbolKind.START, parse_retree.SymbolKind.END)
            ):
                new_concatenants.append(concatenant)

        node.concatenants = new_concatenants
        for concatenant in new_concatenants:
            self.visit(concatenant)


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _translate_pattern(pattern: str) -> Tuple[Optional[str], Optional[str]]:
    """Translate the pattern to obtain the equivalent in XSD."""
    pattern = _undo_escaping_backslash_x_in_pattern(pattern)

    parsed, error = parse_retree.parse(values=[pattern])
    if error is not None:
        regex_line, pointer_line = parse_retree.render_pointer(error.cursor)
        return None, f"{error.message}\n{regex_line}\n{pointer_line}"
    assert parsed is not None

    remover = _AnchorRemover()
    remover.visit(parsed)

    values = parse_retree.render(regex=parsed)
    parts = []  # type: List[str]
    for value in values:
        assert isinstance(value, str), (
            "Only strings expected when rendering a pattern "
            "supplied originally as a string"
        )
        parts.append(value)

    return "".join(parts), None


def _generate_xs_restriction(
    base_type: intermediate.PrimitiveType,
    len_constraint: Optional[infer_for_schema.LenConstraint],
    pattern_constraints: Optional[Sequence[infer_for_schema.PatternConstraint]],
) -> Tuple[Optional[ET.Element], Optional[str]]:
    """
    Generate the ``xs:restriction`` for the given primitive.

    Return the restriction element (if any length or pattern constraints), or
    an error.
    """
    if len_constraint is None and (
        pattern_constraints is None or (len(pattern_constraints) == 0)
    ):
        return None, None

    restriction = ET.Element("xs:restriction", {"base": _PRIMITIVE_MAP[base_type]})

    # NOTE (mristin, 2023-02-27):
    # We skip the general XML character pattern. It makes greenery
    # unbearably slow since it instantiates *each* character in
    # the character range. Since XML engines can not deal with the special
    # characters any ways, there is no need to include this constraint in
    # the XSD pattern restrictions.
    patterns_relevant_for_xsd: Optional[List[infer_for_schema.PatternConstraint]] = None

    if pattern_constraints is not None:
        patterns_relevant_for_xsd = [
            pattern_constraint
            for pattern_constraint in pattern_constraints
            if pattern_constraint.pattern
            != (
                "^[\\x09\\x0A\\x0D\\x20-\\uD7FF\\uE000-\\uFFFD"
                "\\U00010000-\\U0010FFFF]*$"
            )
        ]

    if patterns_relevant_for_xsd is not None and len(patterns_relevant_for_xsd) > 0:
        translated_pattern: Optional[str]

        if len(patterns_relevant_for_xsd) == 1:
            translated_pattern, error = _translate_pattern(
                patterns_relevant_for_xsd[0].pattern
            )
            if error is not None:
                return None, error
        else:
            # NOTE (mristin, 2023-02-27):
            # The module ``greenery`` is not annotated with types at the moment.
            merger = None  # type: Optional[Any]
            for pattern_constraint in patterns_relevant_for_xsd:
                # NOTE (mristin, 2023-02-27):
                # Greenery expects the characters to be in unicode and not escaped.
                translated_for_greenery = _undo_escaping_backslash_x_u_and_U_in_pattern(
                    pattern_constraint.pattern
                )

                try:
                    parsed = greenery.parse(translated_for_greenery)
                except Exception as exception:
                    if translated_for_greenery == pattern_constraint.pattern:
                        return None, (
                            f"The greenery failed to parse "
                            f"the pattern {translated_for_greenery!r}: {exception}"
                        )
                    else:
                        return None, (
                            f"The greenery failed to parse "
                            f"the pattern {translated_for_greenery!r} "
                            f"(which was originally {pattern_constraint.pattern!r}): "
                            f"{exception}"
                        )

                if merger is None:
                    merger = parsed
                else:
                    merger = merger & parsed

            assert merger is not None

            translated_pattern, error = _translate_pattern(str(merger))
            if error is not None:
                return None, error

        assert translated_pattern is not None

        pattern = ET.Element(
            "xs:pattern",
            {"value": translated_pattern},
        )

        restriction.append(pattern)

    if len_constraint is not None:
        if len_constraint.min_value is not None:
            min_length = ET.Element(
                "xs:minLength", {"value": str(len_constraint.min_value)}
            )
            restriction.append(min_length)

        if len_constraint.max_value is not None:
            max_length = ET.Element(
                "xs:maxLength", {"value": str(len_constraint.max_value)}
            )
            restriction.append(max_length)

    return restriction, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_xs_element_for_a_primitive_property(
    prop: intermediate.Property,
    len_constraint: Optional[infer_for_schema.LenConstraint],
    pattern_constraints: Optional[Sequence[infer_for_schema.PatternConstraint]],
) -> Tuple[Optional[ET.Element], Optional[Error]]:
    """
    Generate the ``xs:element`` for a primitive property.

    A primitive property is a property whose type is either a primitive or
    a constrained primitive. The reason why we take these two together is that we
    in-line the constraints for the constrained primitives.

    We do not define the constrained primitives separately in the schema in order to
    avoid the confusion during comparisons between the XSD and the meta-model in
    the book.
    """
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    if (
        isinstance(type_anno, intermediate.OurTypeAnnotation)
        and type_anno.our_type.name == "Value_data_type"
    ):
        # NOTE (mristin, 2022-11-10):
        # Please see :py:const:`_EXPLANATION_ABOUT_WHY_WE_EXPECT_VALUE_DATA_TYPE`
        # for the explanation why we hard-wire the ``Value_data_type`` here
        return (
            ET.Element(
                "xs:element",
                {
                    "name": naming.xml_property(prop.name),
                    "type": "valueDataType",
                },
            ),
            None,
        )

    # NOTE (mristin, 2022-03-30):
    # Specify the type of the ``type_anno`` here with assert instead of specifying it
    # in the pre-condition to help mypy a bit.

    base_type = intermediate.try_primitive_type(type_anno)

    assert (
        base_type is not None
    ), f"Expected a primitive or a constrained primitive, but got: {type_anno}"

    xs_restriction, error = _generate_xs_restriction(
        base_type=base_type,
        len_constraint=len_constraint,
        pattern_constraints=pattern_constraints,
    )
    if error is not None:
        return None, Error(
            prop.parsed.node,
            f"Failed to generate the restriction for property {prop.name}: {error}",
        )
    # NOTE (mristin, 2022-06-18):
    # xs_restriction may be None here if there are no constraints.

    xs_element: ET.Element

    if xs_restriction is None:
        xs_element = ET.Element(
            "xs:element",
            {
                "name": naming.xml_property(prop.name),
                "type": _PRIMITIVE_MAP[base_type],
            },
        )
    else:
        xs_simple_type = ET.Element("xs:simpleType")
        xs_simple_type.append(xs_restriction)

        xs_element = ET.Element("xs:element", {"name": naming.xml_property(prop.name)})
        xs_element.append(xs_simple_type)

    return xs_element, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_xs_element_for_a_list_property(
    prop: intermediate.Property,
    len_constraint: Optional[infer_for_schema.LenConstraint],
) -> Tuple[Optional[ET.Element], Optional[Error]]:
    """Generate the ``xs:element`` for a list property."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    # NOTE (mristin, 2022-03-30):
    # Specify the ``type_anno`` here with assert instead of specifying it
    # in the pre-condition to help mypy a bit
    assert isinstance(type_anno, intermediate.ListTypeAnnotation)

    min_occurs = "0"
    max_occurs = "unbounded"
    if len_constraint is not None:
        if len_constraint.min_value is not None:
            min_occurs = str(len_constraint.min_value)

        if len_constraint.max_value is not None:
            max_occurs = str(len_constraint.max_value)

    xs_element: ET.Element

    if isinstance(type_anno.items, intermediate.OurTypeAnnotation):
        # NOTE (mristin, 2021-11-13):
        # We need to nest the elements in the tag element to separate them in the
        # sequence.

        our_type = type_anno.items.our_type

        if isinstance(our_type, intermediate.Enumeration):
            return None, Error(
                prop.parsed.node,
                f"We do not know how to specify the list of enumerations "
                f"for the property {prop.name!r} of {prop.specified_for.name!r} "
                f"in the XSD with the type: {type_anno}",
            )

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            return None, Error(
                prop.parsed.node,
                f"We do not know how to specify the list of constrained primitives "
                f"for the property {prop.name!r} of {prop.specified_for.name!r} "
                f"in the XSD with the type: {type_anno}",
            )

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            # NOTE (mristin, 2022-05-26):
            # We need to check for the concrete descendants. If there are no concrete
            # descendants, there is no choice group either. Notably, this not only
            # applies to concrete classes, but there is no choice group for the abstract
            # classes without descendants either.
            if len(our_type.concrete_descendants) > 0:
                choice_group_name = xsd_naming.choice_group_name(our_type.name)
                xs_group = ET.Element(
                    "xs:group",
                    {
                        "ref": choice_group_name,
                        "minOccurs": min_occurs,
                        "maxOccurs": max_occurs,
                    },
                )

                xs_sequence = ET.Element("xs:sequence")
                xs_sequence.append(xs_group)

                xs_complex_type = ET.Element("xs:complexType")
                xs_complex_type.append(xs_sequence)

                xs_element = ET.Element(
                    "xs:element", {"name": naming.xml_property(prop.name)}
                )
                xs_element.append(xs_complex_type)
            else:
                xs_element_inner = ET.Element(
                    "xs:element",
                    {
                        "name": naming.xml_class_name(our_type.name),
                        "type": xsd_naming.type_name(our_type.name),
                        "minOccurs": min_occurs,
                        "maxOccurs": max_occurs,
                    },
                )
                xs_sequence = ET.Element("xs:sequence")
                xs_sequence.append(xs_element_inner)

                xs_complex_type = ET.Element("xs:complexType")
                xs_complex_type.append(xs_sequence)

                xs_element = ET.Element(
                    "xs:element", {"name": naming.xml_property(prop.name)}
                )
                xs_element.append(xs_complex_type)

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            return None, Error(
                prop.parsed.node,
                f"We do not know how to specify the list of constrained primitives "
                f"for the property {prop.name!r} of {prop.specified_for.name!r} "
                f"in the XSD with the type: {type_anno}",
            )
        else:
            assert_never(our_type)
    else:
        return None, Error(
            prop.parsed.node,
            f"We do not know how to specify the list "
            f"for the property {prop.name!r} of {prop.specified_for.name!r} "
            f"in the XSD with the type: {type_anno}",
        )

    return xs_element, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_xs_element_for_a_property(
    prop: intermediate.Property,
    len_constraint: Optional[infer_for_schema.LenConstraint],
    pattern_constraints: Optional[Sequence[infer_for_schema.PatternConstraint]],
) -> Tuple[Optional[ET.Element], Optional[Error]]:
    """Generate the definition of an ``xs:element`` for a property."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    xs_element: Optional[ET.Element]

    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
        xs_element, error = _generate_xs_element_for_a_primitive_property(
            prop=prop,
            len_constraint=len_constraint,
            pattern_constraints=pattern_constraints,
        )
        if error is not None:
            return None, error
        assert xs_element is not None

    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        our_type = type_anno.our_type

        if isinstance(our_type, intermediate.Enumeration):
            xs_element = ET.Element(
                "xs:element",
                {
                    "name": naming.xml_property(prop.name),
                    "type": xsd_naming.type_name(our_type.name),
                },
            )

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            xs_element, error = _generate_xs_element_for_a_primitive_property(
                prop=prop,
                len_constraint=len_constraint,
                pattern_constraints=pattern_constraints,
            )
            if error is not None:
                return None, error
            assert xs_element is not None

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            # NOTE (mristin, 2022-05-26):
            # We generate choices only if there are at least one concrete descendant.
            # Otherwise, the choice is not generated. Hence, we need to reference
            # a choice only if there is actually one.
            #
            # This is especially necessary for abstract classes with no descendants
            # which we still want to include in the schema. We simply generate an empty
            # element in the schema for such abstract classes without descendants.
            if len(our_type.concrete_descendants) > 0:
                xs_sequence = ET.Element("xs:sequence")
                xs_sequence.append(
                    ET.Element(
                        "xs:group", {"ref": xsd_naming.choice_group_name(our_type.name)}
                    )
                )

                xs_complex_type = ET.Element("xs:complexType")
                xs_complex_type.append(xs_sequence)

                xs_element = ET.Element(
                    "xs:element", {"name": naming.xml_property(prop.name)}
                )
                xs_element.append(xs_complex_type)
            else:
                xs_element = ET.Element(
                    "xs:element",
                    {
                        "name": naming.xml_property(prop.name),
                        "type": xsd_naming.type_name(our_type.name),
                    },
                )
        else:
            assert_never(type_anno.our_type)

    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        xs_element, error = _generate_xs_element_for_a_list_property(
            prop=prop, len_constraint=len_constraint
        )
        if error is not None:
            return None, error

        assert xs_element is not None
    else:
        assert_never(type_anno)

    assert xs_element is not None

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        xs_element.attrib["minOccurs"] = "0"
        xs_element.attrib["maxOccurs"] = "1"

    return xs_element, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _define_properties(
    cls: intermediate.ClassUnion,
    constraints_by_property: infer_for_schema.ConstraintsByProperty,
) -> Tuple[Optional[List[ET.Element]], Optional[List[Error]]]:
    """Define the properties of the ``cls`` as a sequence of tags."""
    sequence = []  # type: List[ET.Element]
    errors = []  # type: List[Error]

    for prop in cls.properties:
        if prop.specified_for is not cls:
            continue

        len_constraint = constraints_by_property.len_constraints_by_property.get(
            prop, None
        )

        pattern_constraints = constraints_by_property.patterns_by_property.get(
            prop, None
        )

        xs_element, error = _generate_xs_element_for_a_property(
            prop=prop,
            len_constraint=len_constraint,
            pattern_constraints=pattern_constraints,
        )
        if error is not None:
            errors.append(error)
        else:
            assert xs_element is not None
            sequence.append(xs_element)

    if len(errors) > 0:
        return None, errors

    return sequence, None


def _generate_xs_group_for_class(
    cls: intermediate.ClassUnion,
    constraints_by_property: infer_for_schema.ConstraintsByProperty,
) -> Tuple[Optional[ET.Element], Optional[Error]]:
    """Generate the ``xs:group`` representation of the class properties."""
    properties, properties_errors = _define_properties(
        cls=cls, constraints_by_property=constraints_by_property
    )

    if properties_errors is not None:
        return None, Error(
            cls.parsed.node,
            f"Failed to generate xs:group for the class {cls.name!r}",
            properties_errors,
        )

    assert properties is not None

    xs_sequence = ET.Element("xs:sequence")
    for inheritance in cls.inheritances:
        inheritance_xs_group = ET.Element(
            "xs:group", {"ref": xsd_naming.group_name(inheritance.name)}
        )
        xs_sequence.append(inheritance_xs_group)

    xs_sequence.extend(properties)

    xs_group = ET.Element("xs:group", {"name": xsd_naming.group_name(cls.name)})
    xs_group.append(xs_sequence)

    return xs_group, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _define_for_class(
    cls: intermediate.ClassUnion,
    constraints_by_property: infer_for_schema.ConstraintsByProperty,
) -> Tuple[Optional[List[ET.Element]], Optional[Error]]:
    """
    Generate the definitions for the class ``cls``.

    The root element is to be *extended* with the resulting list.
    """
    # NOTE (mristin, 2022-03-30):
    # We define each set of properties in a group. Then we reference these groups
    # among the complex types.
    # See: https://stackoverflow.com/questions/1198755/xml-schemas-with-multiple-inheritance

    xs_group, xs_group_error = _generate_xs_group_for_class(
        cls=cls, constraints_by_property=constraints_by_property
    )
    if xs_group_error is not None:
        return None, xs_group_error

    assert xs_group is not None

    xs_group_ref = ET.Element("xs:group", {"ref": xsd_naming.group_name(cls.name)})

    xs_sequence = ET.Element("xs:sequence")
    xs_sequence.append(xs_group_ref)

    complex_type = ET.Element(
        "xs:complexType", {"name": xsd_naming.type_name(cls.name)}
    )
    complex_type.append(xs_sequence)

    return [xs_group, complex_type], None


@require(lambda cls: len(cls.concrete_descendants) > 0)
def _generate_choice_group(cls: intermediate.ClassUnion) -> ET.Element:
    """Generate a group that defines a choice of concrete descendants."""
    xs_choice = ET.Element("xs:choice")

    if isinstance(cls, intermediate.ConcreteClass):
        xs_choice.append(
            ET.Element(
                "xs:element",
                {
                    "name": naming.xml_class_name(cls.name),
                    "type": xsd_naming.type_name(cls.name),
                },
            )
        )

    for descendant in cls.concrete_descendants:
        xs_choice.append(
            ET.Element(
                "xs:element",
                {
                    "name": naming.xml_class_name(descendant.name),
                    "type": xsd_naming.type_name(descendant.name),
                },
            )
        )

    xs_group = ET.Element("xs:group", {"name": xsd_naming.choice_group_name(cls.name)})
    xs_group.append(xs_choice)
    return xs_group


_WHITESPACE_RE = re.compile(r"\s+")


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _retrieve_implementation_specific_elements(
    cls: intermediate.ClassUnion,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[List[ET.Element]], Optional[List[Error]]]:
    """Parse the elements from the implementation-specific snippet."""
    implementation_key = specific_implementations.ImplementationKey(f"{cls.name}.xml")

    text = spec_impls.get(implementation_key, None)
    if text is None:
        return None, [
            Error(
                cls.parsed.node,
                f"The implementation is missing "
                f"for the implementation-specific class: {implementation_key}",
            )
        ]

    implementation_root: ET.Element

    try:
        implementation_root = ET.fromstring(text)
    except Exception as err:
        return None, [
            Error(
                cls.parsed.node,
                f"Failed to parse the XML out of "
                f"the specific implementation {implementation_key}: {err}",
            )
        ]

    errors = []  # type: List[Error]
    for descendant in implementation_root.iter():
        if descendant.text is not None and not _WHITESPACE_RE.fullmatch(
            descendant.text
        ):
            errors.append(
                Error(
                    cls.parsed.node,
                    f"Unexpected text "
                    f"in the specific implementation {implementation_key} "
                    f"in an element with tag {descendant.tag!r}: {descendant.text!r}",
                )
            )
            continue

        if descendant.tail is not None and not _WHITESPACE_RE.fullmatch(
            descendant.tail
        ):
            errors.append(
                Error(
                    cls.parsed.node,
                    f"Unexpected tail text "
                    f"in the specific implementation {implementation_key} "
                    f"in an element with tag {descendant.tag!r}: {descendant.tail!r}",
                )
            )
            continue

    if len(errors) > 0:
        return None, errors

    # Ignore the implementation root since it defines a partial schema
    elements = []  # type: List[ET.Element]
    for child in implementation_root:
        elements.append(child)

    return elements, None


def _sort_by_tags_and_names_in_place(root: ET.Element) -> None:
    """
    Sort the children elements by tag and name attribute in place.

    This makes diffing and searching in the schema a bit easier.
    """
    groups = []  # type: List[ET.Element]
    simple_types = []  # type: List[ET.Element]
    complex_types = []  # type: List[ET.Element]
    miscellaneous = []  # type: List[ET.Element]
    elements = []  # type: List[ET.Element]

    for child in root:
        if child.tag == "xs:group":
            groups.append(child)
        elif child.tag == "xs:simpleType":
            simple_types.append(child)
        elif child.tag == "xs:complexType":
            complex_types.append(child)
        elif child.tag == "xs:element":
            elements.append(child)
        else:
            miscellaneous.append(child)

    for element_list in [groups, simple_types, complex_types, miscellaneous, elements]:
        element_list.sort(key=lambda elt: elt.attrib.get("name", ""))

    children = groups + simple_types + complex_types + elements + miscellaneous

    assert len(children) == len(root)
    root[:] = children


_EXPLANATION_ABOUT_WHY_WE_EXPECT_VALUE_DATA_TYPE = (
    "(mristin, 2022-09-02) "
    'We provide an internal data type ``valueDataType`` to correspond to "any XSD '
    'atomic type as specified via DataTypeDefXsd". We need this type since we '
    "hard-wire ``Value_data_type`` to it. We could have made "
    "the class ``Value_data_type``implementation-specific and defined its "
    "representation manually as a snippet, including ``valueDataType``.\n\n"
    "However, we decided against that. This would be a major hurdle for "
    "other code and test data generators (which can treat ``Value_data_type`` "
    "simply as string). Therefore, we make the XSD generator "
    "a bit more hacky instead of complicating the other generators.\n\n"
    "If in the future, for whatever reason, the semantic of ``Value_data_type`` "
    "changes (or the type is renamed), be careful to maintain backwards "
    "compatibility here! You probably want to distinguish different versions "
    "of the meta-model and act accordingly. At that point, it might also make "
    "sense to refactor this schema generator to a separate repository, and "
    "fix it to a particular range of meta-model versions."
)


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the XML Schema Definition (XSD) based on the ``symbol_table."""
    root_element_key = specific_implementations.ImplementationKey("root_element.xml")

    root_element_as_text = spec_impls.get(root_element_key, None)
    if root_element_as_text is None:
        return None, [
            Error(
                None,
                f"The implementation snippet for the root element "
                f"is missing: {root_element_key}",
            )
        ]

    root: ET.Element
    try:
        root = ET.fromstring(root_element_as_text)
    except ET.ParseError as err:
        return None, [
            Error(
                None, f"Failed to parse the root element from {root_element_key}: {err}"
            )
        ]

    # NOTE (mristin, 2022-03-30):
    # We need to use minidom to extract the ``xmlns`` property as ElementTree removes
    # it.
    # noinspection PyUnresolvedReferences
    minidom_doc = xml.dom.minidom.parseString(root_element_as_text)

    if not minidom_doc.documentElement.hasAttribute("xmlns"):
        return None, [
            Error(
                None,
                f"The implementation snippet for the root element "
                f"is missing the 'xmlns' attribute: {root_element_key}",
            )
        ]

    xmlns = minidom_doc.documentElement.getAttribute("xmlns")

    if xmlns != symbol_table.meta_model.xml_namespace:
        return None, [
            Error(
                None,
                f"The 'xmlns' attribute of the implementation snippet "
                f"{root_element_key} for the root element "
                f"and the '__xml_namespace__' of the meta-model "
                f"do not coincide: "
                f"{xmlns!r} != {symbol_table.meta_model.xml_namespace!r}",
            )
        ]

    if not minidom_doc.documentElement.hasAttribute("targetNamespace"):
        return None, [
            Error(
                None,
                f"The implementation snippet for the root element "
                f"is missing the 'targetNamespace' attribute: {root_element_key}",
            )
        ]

    target_namespace = minidom_doc.documentElement.getAttribute("targetNamespace")
    if target_namespace != symbol_table.meta_model.xml_namespace:
        return None, [
            Error(
                None,
                f"The 'targetNamespace' attribute of the implementation snippet "
                f"{root_element_key} for the root element "
                f"and the '__xml_namespace__' of the meta-model "
                f"do not coincide: "
                f"{target_namespace!r} != {symbol_table.meta_model.xml_namespace!r}",
            )
        ]

    assert root is not None

    errors = []  # type: List[Error]

    # NOTE (mristin, 2022-04-09):
    # We remove any whitespace tail and text in all the tags, and make sure there is no
    # unexpected text anywhere.
    for element in root.iter():
        if element.text is not None:
            if _WHITESPACE_RE.fullmatch(element.text):
                element.text = None
            else:
                errors.append(
                    Error(
                        None,
                        f"Unexpected text in an element with tag {element.tag!r} "
                        f"from the snippet {root_element_key!r}: {element.text!r}",
                    )
                )

        if element.tail is not None:
            if _WHITESPACE_RE.fullmatch(element.tail):
                element.tail = None
            else:
                errors.append(
                    Error(
                        None,
                        f"Unexpected tail in an element with tag {element.tag!r} "
                        f"from the snippet {root_element_key!r}: {element.tail!r}",
                    )
                )

    if len(errors) > 0:
        return None, errors

    constraints_by_class, some_errors = infer_for_schema.infer_constraints_by_class(
        symbol_table=symbol_table
    )

    if some_errors is not None:
        errors.extend(some_errors)

    # NOTE (mristin, 2022-11-10):
    # Please see :py:const:`_EXPLANATION_ABOUT_WHY_WE_EXPECT_VALUE_DATA_TYPE` for
    # the explanation why we retrieve the ``Value_data_type`` here
    value_data_type_cls = symbol_table.find_our_type(Identifier("Value_data_type"))

    if value_data_type_cls is None:
        errors.append(
            Error(
                None,
                "XSD generator expected to find our type ``Value_data_type``, but it "
                "was not present in the meta-model.\n\n"
                + _EXPLANATION_ABOUT_WHY_WE_EXPECT_VALUE_DATA_TYPE,
            )
        )
    elif not isinstance(value_data_type_cls, intermediate.ConstrainedPrimitive):
        errors.append(
            Error(
                None,
                "XSD generator expected ``Value_data_type`` to be "
                "a constrained primitive,  but got: {type(value_data_type_cls)}.\n\n"
                + _EXPLANATION_ABOUT_WHY_WE_EXPECT_VALUE_DATA_TYPE,
            )
        )
    elif value_data_type_cls.constrainee != intermediate.PrimitiveType.STR:
        errors.append(
            Error(
                None,
                f"XSD generator expected ``Value_data_type`` to be a constrained "
                f"primitive of strings, "
                f"but got: {value_data_type_cls.constrainee}.\n\n"
                + _EXPLANATION_ABOUT_WHY_WE_EXPECT_VALUE_DATA_TYPE,
            )
        )
    else:
        # Our type ``Value_data_type`` is as expected.
        pass

    if len(errors) > 0:
        return None, errors

    assert constraints_by_class is not None

    ids_of_our_types_in_properties = (
        intermediate.collect_ids_of_our_types_in_properties(symbol_table=symbol_table)
    )

    # region Specify ``valueDataType``

    assert value_data_type_cls is not None

    value_data_type_element = ET.Element(
        "xs:simpleType", attrib={"name": "valueDataType"}
    )

    value_data_type_element.append(
        ET.Element(
            "xs:restriction",
            attrib={"base": "xs:string"},
        )
    )

    root.append(value_data_type_element)

    # endregion

    for our_type in symbol_table.our_types:
        if our_type.name == "Value_data_type":
            # NOTE (mristin, 2022-11-10):
            # Please see :py:const:`_EXPLANATION_ABOUT_WHY_WE_EXPECT_VALUE_DATA_TYPE`
            # for the explanation why we hard-wire the ``Value_data_type`` here
            continue

        elements: Optional[List[ET.Element]]

        if (
            isinstance(
                our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
            )
            and our_type.is_implementation_specific
        ):
            elements, impl_spec_errors = _retrieve_implementation_specific_elements(
                cls=our_type, spec_impls=spec_impls
            )
            if impl_spec_errors is not None:
                errors.extend(impl_spec_errors)
                continue

            assert elements is not None
        else:
            if isinstance(our_type, intermediate.Enumeration):
                if id(our_type) not in ids_of_our_types_in_properties:
                    continue

                elements = _define_for_enumeration(enumeration=our_type)

            elif isinstance(our_type, intermediate.ConstrainedPrimitive):
                # NOTE (mristin, 2022-03-30):
                # We in-line the constraints from the constrained primitives directly
                # in the properties. We do not want to introduce separate definitions
                # for them as that would make it more difficult for downstream code
                # generators to generate meaningful code.

                continue

            elif isinstance(
                our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
            ):
                elements, definition_error = _define_for_class(
                    cls=our_type, constraints_by_property=constraints_by_class[our_type]
                )

                if definition_error is not None:
                    errors.append(definition_error)
                    continue

                assert elements is not None

                if len(our_type.concrete_descendants) > 0:
                    choice_group = _generate_choice_group(cls=our_type)
                    elements.append(choice_group)
            else:
                assert_never(our_type)

        assert elements is not None
        root.extend(elements)

    if len(errors) > 0:
        return None, errors

    # Tag name 🠒 (name 🠒 element)
    observed_definitions = dict(
        dict()
    )  # type: MutableMapping[str, MutableMapping[str, ET.Element]]

    for element in root:
        name = element.attrib.get("name", None)
        if name is None:
            continue

        observed_for_tag = observed_definitions.get(element.tag, None)
        if observed_for_tag is None:
            observed_for_tag = dict()
            observed_definitions[element.tag] = observed_for_tag

        observed_element = observed_for_tag.get(name, None)
        if observed_element is not None:
            ours = ET.tostring(element, encoding="unicode", method="xml")
            theirs = ET.tostring(observed_element, encoding="unicode", method="xml")

            errors.append(
                Error(
                    None,
                    f"There are conflicting definitions in the schema "
                    f"with the name {name!r}:\n"
                    f"\n"
                    f"{ours}\n"
                    f"\n"
                    f"and\n"
                    f"\n"
                    f"{theirs}",
                )
            )
        else:
            observed_for_tag[name] = element

    if len(errors) > 0:
        return None, errors

    _sort_by_tags_and_names_in_place(root)

    # NOTE (mristin, 2022-03-30):
    # For some unknown reason, ElementTree erases the xmlns property of the root
    # element. Therefore, we need to add it here manually.
    root.attrib["xmlns"] = xmlns

    text = ET.tostring(root, encoding="unicode", method="xml")

    # NOTE (mristin, 2021-11-23):
    # This approach is slow, but effective. As long as the meta-model is not too big,
    # this should work.
    # noinspection PyUnresolvedReferences
    pretty_text = xml.dom.minidom.parseString(text).toprettyxml(indent="  ")

    return pretty_text, None


def execute(context: run.Context, stdout: TextIO, stderr: TextIO) -> int:
    """Generate the code."""
    code, errors = _generate(
        symbol_table=context.symbol_table, spec_impls=context.spec_impls
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to generate the XML Schema Definition "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert code is not None

    pth = context.output_dir / "schema.xsd"
    try:
        pth.write_text(code, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the XML Schema Definition to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    stdout.write(f"Code generated to: {context.output_dir}\n")
    return 0



"""Generate C++ code of a virtual machine for matching regular expressions."""
import io
from typing import Tuple, Optional, Union, List, Mapping, TextIO

from icontract import ensure, require

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Stripped,
    Identifier,
    indent_but_first_line,
    Error,
    assert_never,
)
from aas_core_codegen.cpp import common as cpp_common, naming as cpp_naming
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
)
from aas_core_codegen.intermediate import revm as intermediate_revm
from aas_core_codegen.parse import retree as parse_retree, tree as parse_tree


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_header(
    symbol_table: intermediate.SymbolTable, library_namespace: Stripped
) -> str:
    """Generate the C++ header of a virtual machine for matching regexes."""
    namespace = Stripped(f"{library_namespace}::{cpp_common.PATTERN_NAMESPACE}")

    fully_qualified_match_function = Stripped(
        f"{library_namespace}::{cpp_common.REVM_NAMESPACE}::Match"
    )

    include_guard_var = cpp_common.include_guard_var(namespace)

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        Stripped(
            f"""\
#ifndef {include_guard_var}
#define {include_guard_var}"""
        ),
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "{include_prefix_path}/revm.hpp"

#pragma warning(push, 0)
#include <memory>
#include <vector>
#pragma warning(pop)"""
        ),
        cpp_common.generate_namespace_opening(library_namespace),
        Stripped(
            f"""\
/**
 * \\defgroup {cpp_common.PATTERN_NAMESPACE} Provide patterns to be matched using a multi-threaded virtual machine.
 *
 * The instructions should be supplied to {fully_qualified_match_function}. While
 * we could have theoretically included this code in verification, we decided to keep
 * it separate for readability. You are not expected to use this module directly.
 * @{{
 */
namespace {cpp_common.PATTERN_NAMESPACE} {{"""
        ),
    ]

    for verification in symbol_table.verification_functions:
        if not isinstance(verification, intermediate.PatternVerification):
            continue

        pattern_name = cpp_naming.constant_name(
            Identifier(f"{verification.name}_program")
        )

        blocks.append(
            Stripped(
                f"""\
extern const std::vector<
{I}std::unique_ptr<revm::Instruction>
> {pattern_name};"""
            )
        )

    blocks.extend(
        [
            Stripped(
                f"""\
}}  // namespace {cpp_common.PATTERN_NAMESPACE}
/**@}}*/"""
            ),
            cpp_common.generate_namespace_closing(library_namespace),
            cpp_common.WARNING,
            Stripped(f"#endif  // {include_guard_var}"),
        ]
    )

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


class _RegexRenderer(parse_retree.Renderer):
    """
    Render the regular expressions for C++ ``std::wregex``.

    In contrast to :py:class:`parse_retree.Renderer`, we also render
    the :py:class:`parse_retree.Char` as we need to cover the granularity of a single
    character in this module.
    """

    def char_to_str_and_escape_or_encode_if_necessary(
        self, node: parse_retree.Char, escaping: Mapping[str, str]
    ) -> List[Union[str, parse_tree.FormattedValue]]:
        """Convert the ``node`` to a string, and escape and/or encode appropriately."""
        if not node.explicitly_encoded:
            escaped = escaping.get(node.character, None)
            if escaped is not None:
                result: List[Union[str, parse_tree.FormattedValue]] = [escaped]
            else:
                result = [node.character]

            return result
        else:
            code = ord(node.character)

            if code <= 255:
                return [f"\\x{code:02x}"]
            elif code <= 65535:
                return [f"\\u{code:04x}"]
            else:
                return [f"\\U{code:08x}"]

    def transform_char(
        self, node: parse_retree.Char
    ) -> List[Union[str, parse_tree.FormattedValue]]:
        return self.char_to_str_and_escape_or_encode_if_necessary(
            node=node, escaping=parse_retree.Renderer._ESCAPING_IN_CHARACTER_LITERALS
        )


_REGEX_RENDERER = _RegexRenderer()


@ensure(lambda result: not result.endswith("\\"))
def _render_comment_re_node(re_node: parse_retree.Node) -> Stripped:
    """Render the pattern as a comment without trailing spaces and line continuation."""
    parts = _REGEX_RENDERER.transform(re_node)
    assert all(
        isinstance(part, str) for part in parts
    ), f"Expected all rendered parts to be strings, but got: {parts}"

    text = "".join(parts)  # type: ignore

    if text.endswith(" ") or text.endswith("\t") or text.endswith("\n"):
        return Stripped(f"/* {text} */")

    if not text.endswith("\\"):
        return Stripped(f"// {text}")

    if "*/" not in text:
        return Stripped(f"/* {text} */")

    return Stripped(
        f"// {text}<PLACEHOLDER TO BREAK LINE CONTINUATION IN C++ COMMENTS>"
    )


@require(lambda indention: indention >= 0)
def _write_instructions_recursively(
    node_or_leaf: intermediate_revm.NodeOrLeaf, indention: int, writer: TextIO
) -> None:
    """Write recursively the creation of the program to the ``writer``."""
    whitespace = I * indention

    if isinstance(node_or_leaf, intermediate_revm.Leaf):
        maybe_label_comment = (
            "" if node_or_leaf.label is None else f"  // {node_or_leaf.label}"
        )

        instruction_code: Stripped

        if isinstance(node_or_leaf.instruction, intermediate_revm.InstructionChar):
            char_literal = cpp_common.wchar_literal(node_or_leaf.instruction.character)
            instruction_code = Stripped(
                f"""\
program.emplace_back({maybe_label_comment}
{I}common::make_unique<revm::InstructionChar>({char_literal})
);"""
            )

        elif isinstance(
            node_or_leaf.instruction,
            (intermediate_revm.InstructionSet, intermediate_revm.InstructionNotSet),
        ):
            ranges = []  # type: List[Stripped]
            for rng in node_or_leaf.instruction.ranges:
                first_literal = cpp_common.wchar_literal(rng.first)
                last_literal = cpp_common.wchar_literal(rng.last)
                ranges.append(Stripped(f"revm::Range({first_literal}, {last_literal})"))

            ranges_joined = ",\n".join(ranges)

            instruction_cls: Stripped
            if isinstance(node_or_leaf.instruction, intermediate_revm.InstructionSet):
                instruction_cls = Stripped("InstructionSet")
            elif isinstance(
                node_or_leaf.instruction, intermediate_revm.InstructionNotSet
            ):
                instruction_cls = Stripped("InstructionNotSet")
            else:
                assert_never(node_or_leaf.instruction)

            instruction_code = Stripped(
                f"""\
program.emplace_back({maybe_label_comment}
{I}common::make_unique<revm::{instruction_cls}>(
{II}std::vector<revm::Range>{{
{III}{indent_but_first_line(ranges_joined, III)}
{II}}}
{I})
);"""
            )

        elif isinstance(node_or_leaf.instruction, intermediate_revm.InstructionAny):
            instruction_code = Stripped(
                f"""\
program.emplace_back({maybe_label_comment}
{I}common::make_unique<revm::InstructionAny>()
);"""
            )

        elif isinstance(node_or_leaf.instruction, intermediate_revm.InstructionMatch):
            instruction_code = Stripped(
                f"""\
program.emplace_back({maybe_label_comment}
{I}common::make_unique<revm::InstructionMatch>()
);"""
            )

        elif isinstance(node_or_leaf.instruction, intermediate_revm.InstructionJump):
            instruction_code = Stripped(
                f"""\
program.emplace_back({maybe_label_comment}
{I}common::make_unique<revm::InstructionJump>({node_or_leaf.instruction.target})
);"""
            )

        elif isinstance(node_or_leaf.instruction, intermediate_revm.InstructionSplit):
            first_target = node_or_leaf.instruction.first_target
            second_target = node_or_leaf.instruction.second_target
            instruction_code = Stripped(
                f"""\
program.emplace_back({maybe_label_comment}
{I}common::make_unique<revm::InstructionSplit>({first_target}, {second_target})
);"""
            )

        elif isinstance(node_or_leaf.instruction, intermediate_revm.InstructionEnd):
            instruction_code = Stripped(
                f"""\
program.emplace_back({maybe_label_comment}
{I}common::make_unique<revm::InstructionEnd>()
);"""
            )

        else:
            assert_never(node_or_leaf.instruction)

        writer.write(
            f"{whitespace}{indent_but_first_line(instruction_code, whitespace)}"
        )

    elif isinstance(node_or_leaf, intermediate_revm.Node):
        re_node_comment = _render_comment_re_node(node_or_leaf.re_node)

        if len(node_or_leaf.children) == 0:
            writer.write(
                f"{whitespace}{{  {re_node_comment}\n"
                f"{whitespace}{I}// Intentionally empty\n"
                f"{whitespace}}}"
            )
        elif len(node_or_leaf.children) == 1 and isinstance(
            node_or_leaf.children[0], intermediate_revm.Leaf
        ):
            writer.write(f"{whitespace}{re_node_comment}\n")
            _write_instructions_recursively(
                node_or_leaf=node_or_leaf.children[0],
                indention=indention,
                writer=writer,
            )
        else:
            writer.write(f"{whitespace}{{  {re_node_comment}\n")

            for child in node_or_leaf.children:
                _write_instructions_recursively(
                    node_or_leaf=child, indention=indention + 1, writer=writer
                )
                writer.write("\n")

            writer.write(f"{whitespace}}}  {re_node_comment}")
    else:
        assert_never(node_or_leaf)


def _transpile_to_instructions(regex: parse_retree.Regex) -> Stripped:
    """
    Transpile the regular expression to instructions of the virtual machine.

    Please see :py:mod:`aas_core_codegen.cpp.revm` for details about the virtual machine
    for matching regular expressions.
    """
    root = intermediate_revm.translate(regex)

    writer = io.StringIO()
    _write_instructions_recursively(node_or_leaf=root, indention=0, writer=writer)

    return Stripped(writer.getvalue())


def _generate_program_definition_for_regex(regex: parse_retree.Regex) -> Stripped:
    """
    Generate the REVM program for the given parsed regular expression.

    We split this function from :py:func:`_generate_construct_function` so that we
    can manually invoke it or debug it on simpler examples.

    The returned block starts with the initialization of ``program`` variable.
    """
    # region Generate everything for the case that wide strings store UTF-32

    instructions_utf32 = _transpile_to_instructions(regex=regex)

    # endregion

    # region Generate everything for the case that wide strings store UTF-16

    parse_retree.fix_for_utf16_regex_in_place(regex=regex)

    instructions_utf16 = _transpile_to_instructions(regex=regex)

    # endregion

    body = [
        Stripped(
            """\
std::vector<std::unique_ptr<revm::Instruction> > program;"""
        )
    ]

    if instructions_utf32 == instructions_utf16:
        body.append(instructions_utf32)
    else:
        body.append(
            Stripped(
                f"""\
#if __WCHAR_MAX__ <= 0x10000
{I}// The size of wchar is 2 bytes.

{I}{indent_but_first_line(instructions_utf16, I)}
#else
{I}// The size of wchar is above 2 bytes.

{I}{indent_but_first_line(instructions_utf32, I)}
#endif"""
            )
        )

    return Stripped("\n\n".join(body))


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_construct_function(
    verification: intermediate.PatternVerification,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the constructor function for the program matching the given pattern."""
    regex, error = parse_retree.parse([verification.pattern])
    if error is not None:
        regex_line, pointer_line = parse_retree.render_pointer(error.cursor)

        return None, Error(
            verification.parsed.node,
            f"Failed to parse the pattern from "
            f"the pattern verification function {verification.name!r}:\n"
            f"{error.message}\n"
            f"{regex_line}\n"
            f"{pointer_line}",
        )

    assert regex is not None

    body = [
        _generate_program_definition_for_regex(regex=regex),
        Stripped("return program;"),
    ]

    body_joined = "\n\n".join(body)

    construct_function = cpp_naming.function_name(
        Identifier(f"construct_{verification.name}_program")
    )

    return (
        Stripped(
            f"""\
std::vector<
{I}std::unique_ptr<revm::Instruction>
> {construct_function}() {{
{I}{indent_but_first_line(body_joined, I)}
}}"""
        ),
        None,
    )


# fmt: off
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(
    symbol_table: intermediate.SymbolTable,
    library_namespace: Stripped,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the C++ implementation of a virtual machine for matching regexes."""
    namespace = Stripped(f"{library_namespace}::{cpp_common.PATTERN_NAMESPACE}")

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f'''\
#include "{include_prefix_path}/common.hpp"
#include "{include_prefix_path}/pattern.hpp"
#include "{include_prefix_path}/revm.hpp"'''
        ),
        cpp_common.generate_namespace_opening(namespace),
    ]

    errors = []  # type: List[Error]

    for verification in symbol_table.verification_functions:
        if not isinstance(verification, intermediate.PatternVerification):
            continue

        block, error = _generate_construct_function(verification=verification)
        if error is not None:
            errors.append(error)
            continue
        else:
            assert block is not None
            blocks.append(block)

        program_name = cpp_naming.constant_name(
            Identifier(f"{verification.name}_program")
        )

        construct_function = cpp_naming.function_name(
            Identifier(f"construct_{verification.name}_program")
        )

        blocks.append(
            Stripped(
                f"""\
const std::vector<
{I}std::unique_ptr<revm::Instruction>
> {program_name} = {construct_function}();"""
            )
        )

    if len(errors) > 0:
        return None, errors

    blocks.extend(
        [
            cpp_common.generate_namespace_closing(namespace),
            cpp_common.WARNING,
        ]
    )

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None



"""
Generate JSON-LD context corresponding to the meta-model.

With the help of the JSON-LD (JSON Linked Data) Context, AAS JSON files can be
represented as triples. That enables applications to make use of Semantic technologies,
such as storing of AAS data in triplestores or -- with the help of appropriate SPARQL
queries -- data transformation, *e.g.*, from AAS Asset Interface Description Submodel
to W3C Web of Things Thing Description.

For more information about JSON-LD, see https://www.w3.org/TR/json-ld11/.

This code has been originally developed by Fabien Amarger (murloc6),
Elodie Thieblin (ethieblin), and Christian Glomb (wiresio).
"""
import collections
import dataclasses
import json
from typing import (
    Any,
    Dict,
    List,
    Set,
    TextIO,
    Optional,
    Tuple,
    Union,
    cast,
    OrderedDict,
)

from icontract import require

from aas_core_codegen import (
    naming,
    intermediate,
    run,
)
from aas_core_codegen.common import Error, Stripped, Identifier, assert_never
from aas_core_codegen.rdf_shacl import (
    naming as rdf_shacl_naming,
    common as rdf_shacl_common,
)

JsonLdType = OrderedDict[str, Any]


def _property_uri(prop: intermediate.Property) -> Stripped:
    """Generate a JSON-LD URI corresponding to the property."""
    property_vocab = rdf_shacl_naming.class_name(prop.specified_for.name)
    prop_uri_fragment = rdf_shacl_naming.property_name(prop.name)

    return Stripped(f"aas:{property_vocab}/{prop_uri_fragment}")


def _get_underlying_atomic_type(
    type_annotation: Union[
        intermediate.PrimitiveTypeAnnotation,
        intermediate.ListTypeAnnotation,
        intermediate.OptionalTypeAnnotation,
        intermediate.PrimitiveTypeAnnotation,
        intermediate.TypeAnnotationUnion,
    ]
) -> Union[intermediate.OurType, intermediate.PrimitiveTypeAnnotation]:
    """Flatten recursively ``our_type`` to obtain the underlying atomic type."""
    if isinstance(type_annotation, intermediate.OptionalTypeAnnotation):
        return _get_underlying_atomic_type(type_annotation.value)
    elif isinstance(type_annotation, intermediate.ListTypeAnnotation):
        return _get_underlying_atomic_type(type_annotation.items)
    elif isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
        return type_annotation
    elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
        return type_annotation.our_type
    else:
        return assert_never(type_annotation)


# fmt: off
@require(
    lambda cls, prop:
    not (cls is not None) or id(prop) in cls.property_id_set
)
# fmt: on
def _generate_for_property(
    cls: Optional[intermediate.ClassUnion],
    prop: intermediate.Property,
    symbol_table: intermediate.SymbolTable,
    our_type_to_rdfs_range: rdf_shacl_common.OurTypeToRdfsRange,
    name_set_of_generated_classes: Optional[Set[str]] = None,
    name_set_of_exported_properties: Optional[Set[str]] = None,
) -> JsonLdType:
    """
    Generate the definition of a property of a class.

    :param cls:
        The class which defines the domain of the property.

        If ``cls`` is None, it means the domain is generic.
    :param prop: The property which will be exported
    :param symbol_table:
        The symbol table corresponding to the AAS meta-model
    :param our_type_to_rdfs_range:
        The mapping between the AAS types and the RDFS range fragment
    :param name_set_of_generated_classes:
        The set of all class names which have been generated so far
    :param name_set_of_exported_properties:
        The set of properties exported so far.

        We specify this set to avoid multiple exports.
    :return: A JSON LD definition of the property
    """
    name_set_of_generated_classes = (
        name_set_of_generated_classes
        if name_set_of_generated_classes is not None
        else set()
    )

    # NOTE (mristin, 2023-10-28):
    # This is necessary for mypy.
    assert name_set_of_generated_classes is not None

    name_set_of_exported_properties = (
        name_set_of_exported_properties
        if name_set_of_exported_properties is not None
        else set()
    )

    # NOTE (mristin, 2023-10-28):
    # This is necessary for mypy.
    assert name_set_of_exported_properties is not None

    property_vocab = rdf_shacl_naming.class_name(prop.specified_for.name)
    prop_uri_fragment = rdf_shacl_naming.property_name(prop.name)

    property_uri = f"aas:{property_vocab}/{prop_uri_fragment}"
    rdfs_range = rdf_shacl_common.rdfs_range_for_type_annotation(
        type_annotation=prop.type_annotation,
        our_type_to_rdfs_range=our_type_to_rdfs_range,
    )
    property_json_ld_context: JsonLdType = collections.OrderedDict(
        [("@id", property_uri)]
    )
    if isinstance(
        intermediate.beneath_optional(prop.type_annotation),
        intermediate.ListTypeAnnotation,
    ):
        property_json_ld_context["@container"] = "@set"

    underlying_atomic_type_annotation = _get_underlying_atomic_type(
        prop.type_annotation
    )

    if isinstance(underlying_atomic_type_annotation, intermediate.Enumeration):
        enum_fragment = rdf_shacl_naming.class_name(
            underlying_atomic_type_annotation.name
        )
        property_json_ld_context["@context"] = cast(
            JsonLdType, collections.OrderedDict([("@vocab", f"aas:{enum_fragment}/")])
        )
        property_json_ld_context["@type"] = "@vocab"

        for item in underlying_atomic_type_annotation.literals:
            rdf_item_name = rdf_shacl_naming.enumeration_literal(item.name)
            json_item_name = item.value
            # Add the property definition if:
            #  * The JSON item name does not start with "xs:" (already defined in XML
            #    Schema Datatypes definition),
            #  * The JSON item name references an already defined class name,
            #  * The JSON item name is different from the RDF item name (we have to
            #    define the correspondance here).
            #
            # Otherwise, adding the "@vocab" is enough.
            if not json_item_name.startswith("xs:") and (
                (json_item_name in name_set_of_generated_classes)
                or json_item_name != rdf_item_name
            ):
                property_json_ld_context["@context"][
                    json_item_name
                ] = collections.OrderedDict(
                    [("@id", f"aas:{enum_fragment}/{rdf_item_name}")]
                )

    elif rdfs_range.startswith("aas:"):
        property_json_ld_context["@type"] = "@id"

        if (
            isinstance(
                underlying_atomic_type_annotation,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            )
            and cls is not underlying_atomic_type_annotation
        ):
            property_json_ld_context["@context"] = collections.OrderedDict()
            for range_property in underlying_atomic_type_annotation.properties:
                range_property_name = naming.json_property(range_property.name)

                if range_property_name not in name_set_of_exported_properties:
                    property_json_ld_context["@context"][
                        range_property_name
                    ] = _generate_for_property(
                        cls=underlying_atomic_type_annotation,
                        prop=range_property,
                        symbol_table=symbol_table,
                        our_type_to_rdfs_range=our_type_to_rdfs_range,
                        name_set_of_generated_classes=name_set_of_generated_classes,
                        name_set_of_exported_properties=name_set_of_exported_properties,
                    )
    elif rdfs_range.startswith("xs:") and rdfs_range not in (
        "xs:string",
        "xs:boolean",
    ):
        property_json_ld_context["@type"] = rdfs_range
    elif rdfs_range == "rdf:langString":
        property_json_ld_context["@container"] = "@set"
        property_json_ld_context["@context"] = collections.OrderedDict(
            [
                ("language", "@language"),
                ("text", "@value"),
            ]
        )
    return property_json_ld_context


def _generate_class_context(
    cls: intermediate.ClassUnion,
    symbol_table: intermediate.SymbolTable,
    our_type_to_rdfs_range: rdf_shacl_common.OurTypeToRdfsRange,
    name_set_of_generated_classes: Optional[Set[str]] = None,
    name_set_of_exported_properties: Optional[Set[str]] = None,
) -> JsonLdType:
    """
    Generate the JSON LD representation for a dedicated class

    :param cls: The dedicated AAS class
    :param symbol_table: The symbol table corresponding to the AAS meta-model
    :param our_type_to_rdfs_range:
        The mapping between the AAS types and the RDFS range fragment
    :param name_set_of_generated_classes:
        The set of all class names which have been generated so far
    :param name_set_of_exported_properties:
        The set of properties exported so far.

         We specify this set to avoid multiple exports.
    :return: The JSON LD representation of the dedicated class

    """
    name_set_of_generated_classes = (
        name_set_of_generated_classes
        if name_set_of_generated_classes is not None
        else set()
    )

    # NOTE (mristin, 2023-10-28):
    # This is necessary for mypy.
    assert name_set_of_generated_classes is not None

    name_set_of_exported_properties = (
        name_set_of_exported_properties
        if name_set_of_exported_properties is not None
        else set()
    )

    # NOTE (mristin, 2023-10-28):
    # This is necessary for mypy.
    assert name_set_of_exported_properties is not None

    class_context_definition: JsonLdType = collections.OrderedDict()
    class_name = naming.json_model_type(cls.name)
    uri_fragment = rdf_shacl_naming.class_name(cls.name)
    class_context_definition[class_name] = collections.OrderedDict(
        [
            ("@id", uri_fragment),
            (
                "@context",
                collections.OrderedDict(
                    [
                        (
                            "@vocab",
                            f"{symbol_table.meta_model.xml_namespace}/{uri_fragment}/",
                        )
                    ]
                ),
            ),
        ]
    )

    for prop in cls.properties:
        property_name = naming.json_property(prop.name)
        if property_name in name_set_of_exported_properties:
            continue

        class_context_definition[class_name]["@context"][
            property_name
        ] = _generate_for_property(
            cls=cls,
            prop=prop,
            symbol_table=symbol_table,
            our_type_to_rdfs_range=our_type_to_rdfs_range,
            name_set_of_generated_classes=name_set_of_generated_classes,
            name_set_of_exported_properties=name_set_of_exported_properties,
        )

    return class_context_definition


@dataclasses.dataclass
class UriAndProperty:
    """Capture a URI for a property along with the property."""

    uri: Stripped
    prop: intermediate.Property


def _generate(
    symbol_table: intermediate.SymbolTable,
    our_type_to_rdfs_range: rdf_shacl_common.OurTypeToRdfsRange,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """
    Generate the JSON-LD context based on the symbol_table.

    :param symbol_table: The symbol table corresponding to the AAS meta-model
    :param our_type_to_rdfs_range:
        The mapping between the AAS types and the RDFS range fragment
    :return: The JSON-LD context as text
    """
    xml_namespace = symbol_table.meta_model.xml_namespace
    json_ld_context: JsonLdType = collections.OrderedDict(
        [
            ("aas", f"{xml_namespace}/"),
            ("xs", "http://www.w3.org/2001/XMLSchema#"),
            ("@vocab", f"{xml_namespace}/"),
            ("modelType", "@type"),
        ]
    )
    errors: List[Error] = []

    class_name_set: Set[str] = set(
        naming.json_model_type(cls.name) for cls in symbol_table.classes
    )

    set_of_property_names_with_double_uris: Set[Identifier] = set()

    uris_and_properties_by_name: Dict[str, UriAndProperty] = dict()

    for cls in symbol_table.classes:
        for prop in cls.properties:
            property_name = naming.json_property(prop.name)
            property_uri = _property_uri(prop)

            if (
                property_name in uris_and_properties_by_name
                and uris_and_properties_by_name[property_name].uri != property_uri
            ):
                set_of_property_names_with_double_uris.add(property_name)

            uris_and_properties_by_name[property_name] = UriAndProperty(
                uri=property_uri, prop=prop
            )

    for property_name in set_of_property_names_with_double_uris:
        del uris_and_properties_by_name[property_name]

    for uri_and_property in uris_and_properties_by_name.values():
        prop = uri_and_property.prop
        property_name = naming.json_property(prop.name)

        if property_name in json_ld_context:
            errors.append(
                Error(
                    prop.parsed.node,
                    f"The property {prop.name!r} will define a duplicate key "
                    f"in JSON-LD Context: {property_name!r}",
                )
            )
            continue

        json_ld_context[property_name] = _generate_for_property(
            cls=None,
            prop=prop,
            symbol_table=symbol_table,
            our_type_to_rdfs_range=our_type_to_rdfs_range,
            name_set_of_generated_classes=class_name_set,
            name_set_of_exported_properties=set(uris_and_properties_by_name.keys()),
        )

    for cls in symbol_table.classes:
        class_context = _generate_class_context(
            cls=cls,
            symbol_table=symbol_table,
            our_type_to_rdfs_range=our_type_to_rdfs_range,
            name_set_of_generated_classes=class_name_set,
            name_set_of_exported_properties=set(uris_and_properties_by_name.keys()),
        )
        if len(class_context.keys() & json_ld_context.keys()) > 0:
            errors.append(
                Error(
                    cls.parsed.node,
                    f"The class {cls.name!r} will have duplicate keys in "
                    f"JSON-LD Context and class context",
                )
            )
        json_ld_context.update(class_context)

    if len(errors) > 0:
        return None, errors

    return Stripped(json.dumps({"@context": json_ld_context}, indent=2)), None


def execute(context: run.Context, stdout: TextIO, stderr: TextIO) -> int:
    """Generate the JSON LD context."""
    our_type_to_rdfs_range, error = rdf_shacl_common.map_our_type_to_rdfs_range(
        symbol_table=context.symbol_table, spec_impls=context.spec_impls
    )
    if error is not None:
        run.write_error_report(
            message=f"Failed to determine the mapping our type to ``rdfs:range`` "
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error)],
            stderr=stderr,
        )

        return 1

    assert our_type_to_rdfs_range is not None

    json_ld_context, errors = _generate(
        symbol_table=context.symbol_table,
        our_type_to_rdfs_range=our_type_to_rdfs_range,
    )

    if errors is not None:
        run.write_error_report(
            message=(
                f"Failed to generate the JSON LD Context based on {context.model_path}"
            ),
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert json_ld_context is not None

    pth = context.output_dir / "context.jsonld"
    try:
        pth.write_text(json_ld_context, encoding="utf-8")
    except Exception as exception:
        run.write_error_report(
            message=f"Failed to write the JSON LD Context to {pth}",
            errors=[str(exception)],
            stderr=stderr,
        )
        return 1

    stdout.write(f"JSON-LD context generated to: {context.output_dir}\n")
    return 0



"""Generate Python code for JSON-ization based on the intermediate representation."""

import io
import textwrap
from typing import Tuple, Optional, List

from icontract import ensure, require

from aas_core_codegen import intermediate, naming, specific_implementations
from aas_core_codegen.common import (
    Error,
    Stripped,
    Identifier,
    assert_never,
    indent_but_first_line,
)
from aas_core_codegen.python import (
    common as python_common,
    naming as python_naming,
)
from aas_core_codegen.python.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# region De-serialization


def _generate_bool_from_jsonable() -> Stripped:
    """Generate the function to decode a ``bool`` from a JSON-able."""
    return Stripped(
        f"""\
def _bool_from_jsonable(
{I}jsonable: Jsonable
) -> bool:
{I}\"\"\"
{I}Parse :paramref:`jsonable` as a boolean.

{I}:param jsonable: JSON-able structure to be parsed
{I}:return: parsed boolean
{I}:raise: :py:class:`DeserializationException` if unexpected :paramref:`jsonable`
{I}\"\"\"
{I}if not isinstance(jsonable, bool):
{II}raise DeserializationException(
{III}f"Expected a bool, but got: {{type(jsonable)}}"
{II})
{I}return jsonable"""
    )


def _generate_int_from_jsonable() -> Stripped:
    """Generate the function to decode an ``int`` from a JSON-able."""
    return Stripped(
        f"""\
def _int_from_jsonable(
{I}jsonable: Jsonable
) -> int:
{I}\"\"\"
{I}Parse :paramref:`jsonable` as an integer.

{I}:param jsonable: JSON-able structure to be parsed
{I}:return: parsed integer
{I}:raise: :py:class:`DeserializationException` if unexpected :paramref:`jsonable`
{I}\"\"\"
{I}if not isinstance(jsonable, int):
{II}raise DeserializationException(
{III}f"Expected an int, but got: {{type(jsonable)}}"
{II})
{I}return jsonable"""
    )


def _generate_float_from_jsonable() -> Stripped:
    """Generate the function to decode a ``float`` from a JSON-able."""
    return Stripped(
        f"""\
def _float_from_jsonable(
{I}jsonable: Jsonable
) -> float:
{I}\"\"\"
{I}Parse :paramref:`jsonable` as a floating-point number.

{I}:param jsonable: JSON-able structure to be parsed
{I}:return: parsed floating-point number
{I}:raise: :py:class:`DeserializationException` if unexpected :paramref:`jsonable`
{I}\"\"\"
{I}if not isinstance(jsonable, float):
{II}raise DeserializationException(
{III}f"Expected a float, but got: {{type(jsonable)}}"
{II})
{I}return jsonable"""
    )


def _generate_str_from_jsonable() -> Stripped:
    """Generate the function to decode a ``str`` from a JSON-able."""
    return Stripped(
        f"""\
def _str_from_jsonable(
{I}jsonable: Jsonable
) -> str:
{I}\"\"\"
{I}Parse :paramref:`jsonable` as a string.

{I}:param jsonable: JSON-able structure to be parsed
{I}:return: parsed string
{I}:raise: :py:class:`DeserializationException` if unexpected :paramref:`jsonable`
{I}\"\"\"
{I}if not isinstance(jsonable, str):
{II}raise DeserializationException(
{III}f"Expected a str, but got: {{type(jsonable)}}"
{II})
{I}return jsonable"""
    )


def _generate_bytes_from_jsonable() -> Stripped:
    """Generate the function to decode ``bytes`` from a JSON-able."""
    return Stripped(
        f"""\
def _bytes_from_jsonable(
{I}jsonable: Jsonable
) -> bytes:
{I}\"\"\"
{I}Decode :paramref:`jsonable` as base64 string to a ``bytearray``.

{I}:param jsonable: JSON-able structure to be decoded
{I}:return: decoded bytearray
{I}:raise: :py:class:`DeserializationException` if unexpected :paramref:`jsonable`
{I}\"\"\"
{I}if not isinstance(jsonable, str):
{II}raise DeserializationException(
{III}f"Expected a str, but got: {{type(jsonable)}}"
{II})

{I}return base64.b64decode(
{II}jsonable.encode('ascii')
{I})"""
    )


def _generate_enumeration_from_jsonable(
    enumeration: intermediate.Enumeration,
) -> Stripped:
    """Generate the deserialization method for an enumeration."""
    enum_name = python_naming.enum_name(identifier=enumeration.name)

    function_name = python_naming.function_name(
        Identifier(f"{enumeration.name}_from_jsonable")
    )

    enum_from_str = python_naming.function_name(
        Identifier(f"{enumeration.name}_from_str")
    )

    return Stripped(
        f"""\
def {function_name}(
{I}jsonable: Jsonable
) -> aas_types.{enum_name}:
{I}\"\"\"
{I}Convert the JSON-able structure :paramref:`jsonable` to a literal of
{I}:py:class:`.types.{enum_name}`.

{I}:param jsonable: JSON-able structure to be parsed
{I}:return: parsed literal
{I}:raise: :py:class:`.DeserializationException` if unexpected :paramref:`jsonable`
{I}\"\"\"
{I}if not isinstance(jsonable, str):
{II}raise DeserializationException(
{III}"Expected a str, but got: {{type(jsonable)}}"
{II})

{I}literal = aas_stringification.{enum_from_str}(jsonable)
{I}if literal is None:
{II}raise DeserializationException(
{III}f"Not a valid string representation of "
{III}f"a literal of {enum_name}: {{jsonable}}"
{II})

{I}return literal"""
    )


def _generate_is_array_like() -> Stripped:
    """Generate the function to check that the jsonable is an array-like object."""
    return Stripped(
        f'''\
def _try_to_cast_to_array_like(
{I}jsonable: Jsonable
) -> Optional[Iterable[Any]]:
{I}"""
{I}Try to cast the ``jsonable`` to something like a JSON array.

{I}In particular, we explicitly check that the ``jsonable`` is not a mapping, as we
{I}do not want to mistake dictionaries (*i.e.* de-serialized JSON objects) for lists.

{I}>>> assert _try_to_cast_to_array_like(True) is None

{I}>>> assert _try_to_cast_to_array_like(0) is None

{I}>>> assert _try_to_cast_to_array_like(2.2) is None

{I}>>> assert _try_to_cast_to_array_like("hello") is None

{I}>>> assert _try_to_cast_to_array_like(b"hello") is None

{I}>>> _try_to_cast_to_array_like([1, 2])
{I}[1, 2]

{I}>>> assert _try_to_cast_to_array_like({{"a": 3}}) is None

{I}>>> assert _try_to_cast_to_array_like(collections.OrderedDict()) is None

{I}>>> _try_to_cast_to_array_like(range(1, 2))
{I}range(1, 2)

{I}>>> _try_to_cast_to_array_like((1, 2))
{I}(1, 2)

{I}>>> assert _try_to_cast_to_array_like({{1, 2, 3}}) is None
{I}"""
{I}if (
{II}not isinstance(jsonable, (str, bytearray, bytes))
{II}and hasattr(jsonable, "__iter__")
{II}and not hasattr(jsonable, "keys")
{II}# NOTE (mristin):
{II}# There is no easy way to check for sets as opposed to sequence except
{II}# for checking for direct inheritance. A sequence also inherits from
{II}# a collection, so both sequences and sets provide ``__contains__`` method.
{II}#
{II}# See: https://docs.python.org/3/library/collections.abc.html
{II}and not isinstance(jsonable, collections.abc.Set)
{I}):
{II}return cast(Iterable[Any], jsonable)

{I}return None'''
    )


def _generate_dispatch_map_for_abstract_class(
    cls: intermediate.AbstractClass,
) -> Stripped:
    """Generate a mapping model type 🠒 de-serialization function."""
    assert len(cls.concrete_descendants) > 0, (
        "Expected an abstract class to have concrete descendants. "
        "Otherwise we do not know how to de-serialize it."
    )

    mapping_name = python_naming.private_constant_name(
        Identifier(f"{cls.name}_from_jsonable_dispatch")
    )

    cls_name = python_naming.class_name(cls.name)

    mapping_writer = io.StringIO()
    mapping_writer.write(
        f"""\
{mapping_name}: Mapping[
{I}str,
{I}Callable[[Jsonable], aas_types.{cls_name}]
] = {{
"""
    )

    for descendant in cls.concrete_descendants:
        function_name_for_descendant = python_naming.function_name(
            Identifier(f"{descendant.name}_from_jsonable")
        )

        descendant_literal = python_common.string_literal(
            naming.json_model_type(descendant.name)
        )

        mapping_writer.write(
            f"""\
{I}{descendant_literal}: {function_name_for_descendant},
"""
        )

    mapping_writer.write("}")

    return Stripped(mapping_writer.getvalue())


@require(lambda cls: len(cls.concrete_descendants) > 0)
def _generate_dispatch_map_for_concrete_class(
    cls: intermediate.ConcreteClass,
) -> Stripped:
    """
    Generate a mapping model type 🠒 de-serialization function.

    The ``jsonable`` can not be parsed straight-away as we don't know which class
    (the actual one or one of the descendants) needs to be parsed.
    """
    mapping_name = python_naming.private_constant_name(
        Identifier(f"{cls.name}_from_jsonable_dispatch")
    )

    cls_name = python_naming.class_name(cls.name)

    mapping_writer = io.StringIO()
    mapping_writer.write(
        f"""\
{mapping_name}: Mapping[
{I}str,
{I}Callable[[Jsonable], aas_types.{cls_name}]
] = {{
"""
    )

    function_name_for_cls = python_naming.function_name(
        Identifier(f"_{cls.name}_from_jsonable_without_dispatch")
    )

    cls_literal = python_common.string_literal(naming.json_model_type(cls.name))

    mapping_writer.write(
        f"""\
{I}{cls_literal}: {function_name_for_cls},
"""
    )

    for descendant in cls.concrete_descendants:
        function_name_for_descendant = python_naming.function_name(
            Identifier(f"{descendant.name}_from_jsonable")
        )

        descendant_literal = python_common.string_literal(
            naming.json_model_type(descendant.name)
        )

        mapping_writer.write(
            f"""\
{I}{descendant_literal}: {function_name_for_descendant},
"""
        )

    mapping_writer.write("}")

    return Stripped(mapping_writer.getvalue())


# fmt: off
@require(
    lambda cls: len(cls.concrete_descendants) > 0,
    "Dispatch is only possible if there are one or more concrete descendants"
)
# fmt: on
def _generate_dispatch_from_jsonable(cls: intermediate.ClassUnion) -> Stripped:
    """Generate the de-serialization dispatch for an abstract class."""
    function_name = python_naming.function_name(Identifier(f"{cls.name}_from_jsonable"))

    cls_name = python_naming.class_name(cls.name)

    mapping_name = python_naming.private_constant_name(
        Identifier(f"{cls.name}_from_jsonable_dispatch")
    )

    return Stripped(
        f"""\
def {function_name}(
{II}jsonable: Jsonable
) -> aas_types.{cls_name}:
{I}\"\"\"
{I}Parse an instance of :py:class:`.types.{cls_name}` from the JSON-able
{I}structure :paramref:`jsonable`.

{I}:param jsonable: structure to be parsed
{I}:return: Concrete instance of :py:class:`.types.{cls_name}`
{I}:raise: :py:class:`DeserializationException` if unexpected :paramref:`jsonable`
{I}\"\"\"
{I}if not isinstance(jsonable, collections.abc.Mapping):
{II}raise DeserializationException(
{III}f"Expected a mapping, but got: {{type(jsonable)}}"
{II})

{I}model_type = jsonable.get("modelType", None)
{I}if model_type is None:
{II}raise DeserializationException(
{III}"Expected the property modelType, but found none"
{II})

{I}if not isinstance(model_type, str):
{II}raise DeserializationException(
{III}"Expected the property modelType to be a str, but got: {{type(model_type)}}"
{II})

{I}dispatch = {mapping_name}.get(model_type, None)
{I}if dispatch is None:
{II}raise DeserializationException(
{III}f"Unexpected model type for {cls_name}: {{model_type}}"
{II})

{I}return dispatch(jsonable)"""
    )


_PARSE_FUNCTION_BY_PRIMITIVE_TYPE = {
    intermediate.PrimitiveType.BOOL: "_bool_from_jsonable",
    intermediate.PrimitiveType.INT: "_int_from_jsonable",
    intermediate.PrimitiveType.FLOAT: "_float_from_jsonable",
    intermediate.PrimitiveType.STR: "_str_from_jsonable",
    intermediate.PrimitiveType.BYTEARRAY: "_bytes_from_jsonable",
}
assert all(
    literal in _PARSE_FUNCTION_BY_PRIMITIVE_TYPE
    for literal in intermediate.PrimitiveType
)


def _parse_function_for_atomic_value(
    type_annotation: intermediate.AtomicTypeAnnotation,
) -> Stripped:
    """Determine the parse function for deserializing an atomic non-optional value."""
    function_name: str

    if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
        function_name = _PARSE_FUNCTION_BY_PRIMITIVE_TYPE[type_annotation.a_type]

    elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
        our_type = type_annotation.our_type

        if isinstance(
            our_type,
            (
                intermediate.Enumeration,
                intermediate.AbstractClass,
                intermediate.ConcreteClass,
            ),
        ):
            function_name = python_naming.function_name(
                Identifier(f"{our_type.name}_from_jsonable")
            )

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            function_name = _PARSE_FUNCTION_BY_PRIMITIVE_TYPE[our_type.constrainee]

        else:
            assert_never(our_type)
    else:
        assert_never(type_annotation)

    return Stripped(function_name)


def _generate_setter(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate a class which allows us to dispatch de-serialization for properties."""
    methods = []  # type: List[Stripped]

    init_writer = io.StringIO()
    for i, prop in enumerate(cls.properties):
        prop_name = python_naming.property_name(prop.name)
        prop_type = python_common.generate_type(
            prop.type_annotation, types_module=Identifier("aas_types")
        )

        # NOTE (mristin, 2022-07-22):
        # We make all the properties optional since we switch over the properties
        # during the de-serialization.
        if not isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            prop_type = Stripped(f"Optional[{prop_type}]")

        if i > 0:
            init_writer.write("\n")
        init_writer.write(f"self.{prop_name}: {prop_type} = None")

    methods.append(
        Stripped(
            f"""\
def __init__(self) -> None:
{I}\"\"\"Initialize with all the properties unset.\"\"\"
{I}{indent_but_first_line(init_writer.getvalue(), I)}"""
        )
    )

    methods.append(
        Stripped(
            f"""\
def ignore(self, jsonable: Jsonable) -> None:
{I}\"\"\"Ignore :paramref:`jsonable` and do not set anything.\"\"\"
{I}pass"""
        )
    )

    for i, prop in enumerate(cls.properties):
        prop_name = python_naming.property_name(prop.name)

        method_name = python_naming.method_name(
            Identifier(f"set_{prop.name}_from_jsonable")
        )

        type_anno = intermediate.beneath_optional(prop.type_annotation)
        if isinstance(
            type_anno,
            (intermediate.PrimitiveTypeAnnotation, intermediate.OurTypeAnnotation),
        ):
            function_name = _parse_function_for_atomic_value(type_anno)
            body = Stripped(
                f"""\
self.{prop_name} = {function_name}(
{I}jsonable
)"""
            )

        elif isinstance(type_anno, intermediate.ListTypeAnnotation):
            assert not isinstance(
                type_anno.items,
                (intermediate.OptionalTypeAnnotation, intermediate.ListTypeAnnotation),
            ), (
                "We chose to implement only a very limited pattern matching; "
                "see intermediate._translate_._verify_only_simple_type_patterns"
            )

            items_type = python_common.generate_type(
                type_anno.items, types_module=Identifier("aas_types")
            )
            parse_function = _parse_function_for_atomic_value(type_anno.items)

            body = Stripped(
                f"""\
array_like = _try_to_cast_to_array_like(jsonable)
if array_like is None:
{I}raise DeserializationException(
{II}f"Expected something array-like, but got: {{type(jsonable)}}"
{I})

items: List[
{I}{items_type}
] = []
for i, jsonable_item in enumerate(array_like):
{I}try:
{II}item = {parse_function}(
{III}jsonable_item
{II})
{I}except DeserializationException as exception:
{II}exception.path._prepend(
{III}IndexSegment(
{IIII}array_like,
{IIII}i
{III})
{II})
{II}raise

{I}items.append(item)

self.{prop_name} = items"""
            )

        else:
            assert_never(type_anno)

        method_writer = io.StringIO()
        method_writer.write(
            f"""\
def {method_name}(
{II}self,
{II}jsonable: Jsonable
) -> None:
{I}\"\"\"
{I}Parse :paramref:`jsonable` as the value of :py:attr:`~{prop_name}`.

{I}:param jsonable: input to be parsed
{I}\"\"\"
{I}{indent_but_first_line(body, I)}"""
        )

        methods.append(Stripped(method_writer.getvalue()))

    cls_name = python_naming.private_class_name(Identifier(f"Setter_for_{cls.name}"))

    writer = io.StringIO()
    writer.write(
        f"""\
class {cls_name}:
{I}\"\"\"Provide de-serialization-setters for properties.\"\"\"

"""
    )

    assert len(methods) > 0, (
        "Expected at least one method, otherwise the newlines are invalid and "
        "have to be fixed"
    )

    for i, method in enumerate(methods):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(method, I))

    return Stripped(writer.getvalue())


def _generate_setter_map(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate a mapping ``JSON property name`` -> deserialization on a setter."""
    # fmt: off
    assert (
            sorted(
                (arg.name, str(arg.type_annotation))
                for arg in cls.constructor.arguments
            ) == sorted(
                (prop.name, str(prop.type_annotation))
                for prop in cls.properties
            )
    ), (
        "(mristin, 2022-10-03) We assume that the properties and constructor arguments "
        "are identical at this point. If this is not the case, we have to re-write the "
        "logic substantially! Please contact the developers if you see this."
    )
    # fmt: on

    identifiers_expressions = []  # type: List[Tuple[Identifier, Stripped]]

    setter_cls_name = python_naming.private_class_name(
        Identifier(f"Setter_for_{cls.name}")
    )

    for prop in cls.properties:
        json_identifier = naming.json_property(prop.name)
        method_name = python_naming.method_name(
            Identifier(f"set_{prop.name}_from_jsonable")
        )

        identifiers_expressions.append(
            (json_identifier, Stripped(f"{setter_cls_name}.{method_name}"))
        )

    map_name = python_naming.private_constant_name(
        Identifier(f"setter_map_for_{cls.name}")
    )

    writer = io.StringIO()
    writer.write(
        f"""\
{map_name}: Mapping[
{I}str,
{I}Callable[
{II}[{setter_cls_name}, Jsonable],
{II}None
{I}]
] = {{
"""
    )
    for identifier, expression in identifiers_expressions:
        writer.write(
            f"""\
{I}{python_common.string_literal(identifier)}:
{II}{indent_but_first_line(expression, II)},
"""
        )

    writer.write(
        f"""\
{I}'modelType':
{II}{setter_cls_name}.ignore
"""
    )

    writer.write("}")
    return Stripped(writer.getvalue())


def _generate_concrete_class_from_jsonable(
    cls: intermediate.ConcreteClass,
) -> Stripped:
    """
    Generate the deserialization function for a concrete class.

    This function performs no dispatch. If it de-serializes a concrete class with
    concrete descendants, we have to provide a different name. Otherwise, it would
    shadow the name for the dispatch function.
    """
    # fmt: off
    assert (
            sorted(
                (arg.name, str(arg.type_annotation))
                for arg in cls.constructor.arguments
            ) == sorted(
                (prop.name, str(prop.type_annotation))
                for prop in cls.properties
            )
    ), (
        "(mristin, 2022-10-03) We assume that the properties and constructor arguments "
        "are identical at this point. If this is not the case, we have to re-write the "
        "logic substantially! Please contact the developers if you see this."
    )
    # fmt: on

    setter_cls_name = python_naming.private_class_name(
        Identifier(f"Setter_for_{cls.name}")
    )

    blocks = [
        Stripped(
            f"""\
if not isinstance(jsonable, collections.abc.Mapping):
{I}raise DeserializationException(
{II}f"Expected a mapping, but got: {{type(jsonable)}}"
{I})"""
        ),
        Stripped(f"setter = {setter_cls_name}()"),
    ]  # type: List[Stripped]

    # NOTE (mristin):
    # If the class has no concrete descendants, this function will be the sole entry
    # point for the de-serialization. In that case, we have to explicitly check if
    # the ``modelType`` is present — for cases where the serialization rules impose
    # a model type for backwards compatibility even though it is redundant.
    #
    # If the class has concrete descendants, we can only publicly de-serialize it
    # through a dispatching function, which will innately check for model type, so we
    # do not have to repeat the check here.
    if len(cls.concrete_descendants) == 0 and cls.serialization.with_model_type:
        expected_model_type = naming.json_model_type(cls.name)
        blocks.append(
            Stripped(
                f"""\
model_type = jsonable.get("modelType", None)
if model_type is None:
{I}raise DeserializationException(
{II}"Expected the property modelType, but found none"
{I})

if model_type != {python_common.string_literal(expected_model_type)}:
{I}raise DeserializationException(
{II}f"Invalid modelType, expected '{expected_model_type}', "
{II}f"but got: {{model_type!r}}"
{I})"""
            )
        )

    # region Switch on property name

    map_name = python_naming.private_constant_name(
        Identifier(f"setter_map_for_{cls.name}")
    )

    blocks.append(
        Stripped(
            f"""\
for key, jsonable_value in jsonable.items():
{I}setter_method = (
{II}{map_name}.get(key)
{I})
{I}if setter_method is None:
{II}raise DeserializationException(
{III}f"Unexpected property: {{key}}"
{II})

{I}try:
{II}setter_method(setter, jsonable_value)
{I}except DeserializationException as exception:
{II}exception.path._prepend(
{III}PropertySegment(
{IIII}jsonable_value,
{IIII}key
{III})
{II})
{II}raise exception"""
        )
    )

    # region Check required properties

    required_checks = []  # type: List[Stripped]
    for i, prop in enumerate(cls.properties):
        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            continue

        prop_name = python_naming.property_name(prop.name)

        cause_literal = python_common.string_literal(
            f"The required property {naming.json_property(prop.name)!r} is missing"
        )
        required_checks.append(
            Stripped(
                f"""\
if setter.{prop_name} is None:
{I}raise DeserializationException(
{II}{cause_literal}
{I})"""
            )
        )

    if len(required_checks) > 0:
        blocks.append(Stripped("\n\n".join(required_checks)))

    # endregion

    # region Pass in arguments to the constructor

    cls_name = python_naming.class_name(cls.name)

    if len(cls.constructor.arguments) == 0:
        blocks.append(Stripped(f"return aas_types.{cls_name}()"))
    else:
        init_writer = io.StringIO()
        init_writer.write(f"return aas_types.{cls_name}(\n")

        for i, arg in enumerate(cls.constructor.arguments):
            prop = cls.properties_by_name[arg.name]

            prop_name = python_naming.property_name(prop.name)

            init_writer.write(f"{I}setter.{prop_name}")

            if i < len(cls.constructor.arguments) - 1:
                init_writer.write(",\n")
            else:
                init_writer.write("\n")

        init_writer.write(")")

        blocks.append(Stripped(init_writer.getvalue()))
    # endregion

    writer = io.StringIO()

    if len(cls.concrete_descendants) == 0:
        function_name = python_naming.function_name(
            Identifier(f"{cls.name}_from_jsonable")
        )
    else:
        function_name = python_naming.function_name(
            Identifier(f"_{cls.name}_from_jsonable_without_dispatch")
        )

    docstring_blocks = [
        Stripped(
            f"""\
Parse an instance of :py:class:`.types.{cls_name}` from the JSON-able
structure :paramref:`jsonable`."""
        )
    ]  # type: List[Stripped]

    if len(cls.concrete_descendants) > 0:
        function_name_with_dispatch = python_naming.function_name(
            Identifier(f"{cls.name}_from_jsonable")
        )

        docstring_blocks.append(
            Stripped(
                f"""\
This function performs no dispatch! It is used to parse the properties
as-are, and already assumes the exact model type. Usually, this function
is called from within a dispatching function, and you never call it
directly. If you want to de-serialize an instance of
:py:class:`.types.{cls_name}`, call
:py:func:`{function_name_with_dispatch}`."""
            )
        )

    docstring_blocks.append(
        Stripped(
            f"""\
:param jsonable: structure to be parsed
:return: Parsed instance of :py:class:`.types.{cls_name}`
:raise: :py:class:`DeserializationException` if unexpected :paramref:`jsonable`"""
        )
    )

    docstring = "\n\n".join(docstring_blocks)

    writer.write(
        f"""\
def {function_name}(
{II}jsonable: Jsonable
) -> aas_types.{cls_name}:
{I}\"\"\"
{I}{indent_but_first_line(docstring, I)}
{I}\"\"\"
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    return Stripped(writer.getvalue())


# endregion

# region Serialization


def _generate_bytes_to_base64_str() -> Stripped:
    """Generate the function to encode bytes as base64-encoded string."""
    return Stripped(
        f"""\
def _bytes_to_base64_str(
{I}value: bytes
) -> str:
{I}\"\"\"
{I}Encode :paramref:`value` as a base64 string.

{I}:param value: to be encoded
{I}:return: encoded :paramref:`value` in base64
{I}\"\"\"
{I}# We need to decode as ascii as ``base64.b64encode`` returns bytes,
{I}# not a string!
{I}return base64.b64encode(value).decode('ascii')"""
    )


def _generate_transform_atomic_value(
    access_expression: str, type_anno: intermediate.AtomicTypeAnnotation
) -> Stripped:
    """
    Generate the snippet to transform the ``access_expression``.

    The ``access_expression`` should either be a name or a member access.
    """
    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation) or (
        isinstance(type_anno, intermediate.OurTypeAnnotation)
        and isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive)
    ):
        a_type = intermediate.try_primitive_type(type_anno)
        assert a_type is not None

        if (
            a_type is intermediate.PrimitiveType.BOOL
            or a_type is intermediate.PrimitiveType.INT
            or a_type is intermediate.PrimitiveType.FLOAT
            or a_type is intermediate.PrimitiveType.STR
        ):
            return Stripped(f"{access_expression}")

        elif a_type is intermediate.PrimitiveType.BYTEARRAY:
            return Stripped(f"_bytes_to_base64_str({access_expression})")

        else:
            assert_never(a_type)

    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        if isinstance(type_anno.our_type, intermediate.Enumeration):
            return Stripped(f"{access_expression}.value")

        elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
            raise AssertionError("This case should have been handled before.")

        elif isinstance(
            type_anno.our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            return Stripped(f"self.transform({access_expression})")

        else:
            assert_never(type_anno.our_type)

    else:
        assert_never(type_anno.our_type)


def _generate_transform(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the ``transform_X`` method to serialize an instance into a JSON-able."""
    blocks = [
        Stripped("jsonable: MutableMapping[str, MutableJsonable] = dict()")
    ]  # type: List[Stripped]

    for prop in cls.properties:
        key_literal = python_common.string_literal(naming.json_property(prop.name))
        prop_name = python_naming.property_name(prop.name)

        type_anno = intermediate.beneath_optional(prop.type_annotation)

        block: Stripped

        if isinstance(
            type_anno,
            (intermediate.PrimitiveTypeAnnotation, intermediate.OurTypeAnnotation),
        ):
            transformation_expression = _generate_transform_atomic_value(
                access_expression=Stripped(f"that.{prop_name}"), type_anno=type_anno
            )

            block = Stripped(f"jsonable[{key_literal}] = {transformation_expression}")

            # Rudimentary formatting heuristics
            if len(block) > 70:
                block = Stripped(
                    f"""\
jsonable[{key_literal}] = (
{I}{transformation_expression}
)"""
                )

        elif isinstance(type_anno, intermediate.ListTypeAnnotation):
            assert isinstance(
                type_anno.items,
                (intermediate.PrimitiveTypeAnnotation, intermediate.OurTypeAnnotation),
            ), (
                "We expect only lists of primitive and our types. Lists of optionals "
                "and nested lists are not handled yet. Please contact the developers."
            )

            transformation_expression = _generate_transform_atomic_value(
                access_expression=Stripped("item"), type_anno=type_anno.items
            )

            block = Stripped(
                f"""\
jsonable[{key_literal}] = [
{I}{transformation_expression}
{I}for item in that.{prop_name}
]"""
            )

        else:
            assert_never(type_anno)

        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            block = Stripped(
                f"""\
if that.{prop_name} is not None:
{I}{indent_but_first_line(block, I)}"""
            )

        blocks.append(block)

    if cls.serialization.with_model_type:
        model_type_literal = python_common.string_literal(
            naming.json_model_type(cls.name)
        )
        blocks.append(Stripped(f"""jsonable["modelType"] = {model_type_literal}"""))

    blocks.append(Stripped("return jsonable"))

    method_name = python_naming.method_name(Identifier(f"transform_{cls.name}"))

    cls_name = python_naming.class_name(cls.name)

    writer = io.StringIO()

    no_self_use = all(
        (
            some_type_anno := intermediate.beneath_optional(prop.type_annotation),
            intermediate.try_primitive_type(some_type_anno) is not None
            or (
                isinstance(some_type_anno, intermediate.OurTypeAnnotation)
                and isinstance(some_type_anno.our_type, intermediate.Enumeration)
            ),
        )[1]
        for prop in cls.properties
    )

    if no_self_use:
        writer.write("# noinspection PyMethodMayBeStatic\n")

    writer.write(
        f"""\
def {method_name}(
{I}self,
{I}that: aas_types.{cls_name}
) -> MutableJsonable:
{I}\"\"\"Serialize :paramref:`that` to a JSON-able representation.\"\"\""""
    )

    if len(blocks) > 0:
        writer.write("\n")

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    return Stripped(writer.getvalue())


def _generate_transformer(symbol_table: intermediate.SymbolTable) -> Stripped:
    methods = []  # type: List[Stripped]

    for our_type in symbol_table.our_types:
        if isinstance(our_type, intermediate.ConcreteClass):
            methods.append(_generate_transform(our_type))

    writer = io.StringIO()
    writer.write(
        f"""\
class _Serializer(
{II}aas_types.AbstractTransformer[MutableJsonable]
):
{I}\"\"\"Transform the instance to its JSON-able representation.\"\"\""""
    )

    for method in methods:
        writer.write("\n\n")
        writer.write(textwrap.indent(method, I))

    return Stripped(writer.getvalue())


# endregion

# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    qualified_module_name: python_common.QualifiedModuleName,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the Python code for the general serialization.

    The ``qualified_module_name`` indicates the fully-qualified name of the base module.
    """
    blocks = [
        Stripped(
            """\
\"\"\"
Provide de/serialization of AAS classes to/from JSON.

We can not use one-pass deserialization for JSON since the object
properties do not have fixed order, and hence we can not read
``modelType`` property ahead of the remaining properties.
\"\"\""""
        ),
        python_common.WARNING,
        Stripped(
            f"""\
import base64
import collections.abc
import sys
from typing import (
{I}cast,
{I}Any,
{I}Callable,
{I}Iterable,
{I}List,
{I}Mapping,
{I}MutableMapping,
{I}Optional,
{I}Sequence,
{I}Union,
)

if sys.version_info >= (3, 8):
{I}from typing import Final
else:
{I}from typing_extensions import Final

import {qualified_module_name}.common as aas_common
import {qualified_module_name}.stringification as aas_stringification
import {qualified_module_name}.types as aas_types"""
        ),
        Stripped(
            f"""\
class PropertySegment:
{I}\"\"\"Represent a property on a path to the erroneous value.\"\"\"

{I}#: Instance that contains the property
{I}instance: Final[Mapping[str, Any]]

{I}#: Name of the property
{I}name: Final[str]

{I}def __init__(
{III}self,
{III}instance: Mapping[str, Any],
{III}name: str
{I}) -> None:
{II}\"\"\"Initialize with the given values.\"\"\"
{II}self.instance = instance
{II}self.name = name"""
        ),
        Stripped(
            f"""\
class IndexSegment:
{I}\"\"\"Represent an index access on a path to the erroneous value.\"\"\"

{I}#: Container that contains the item
{I}container: Final[Iterable[Any]]

{I}#: Index of the item
{I}index: Final[int]

{I}def __init__(
{III}self,
{III}container: Iterable[Any],
{III}index: int
{I}) -> None:
{II}\"\"\"Initialize with the given values.\"\"\"
{II}self.container = container
{II}self.index = index"""
        ),
        Stripped(
            """\
Segment = Union[PropertySegment, IndexSegment]"""
        ),
        Stripped(
            f"""\
class Path:
{I}\"\"\"Represent the relative path to the erroneous value.\"\"\"

{I}def __init__(self) -> None:
{II}\"\"\"Initialize as an empty path.\"\"\"
{II}self._segments = []  # type: List[Segment]

{I}@property
{I}def segments(self) -> Sequence[Segment]:
{II}\"\"\"Get the segments of the path.\"\"\"
{II}return self._segments

{I}def _prepend(self, segment: Segment) -> None:
{II}\"\"\"Insert the :paramref:`segment` in front of other segments.\"\"\"
{II}self._segments.insert(0, segment)

{I}def __str__(self) -> str:
{II}if len(self._segments) == 0:
{III}return ""

{II}parts = []  # type: List[str]

{II}iterator = iter(self._segments)
{II}first = next(iterator)
{II}if isinstance(first, PropertySegment):
{III}parts.append(f"{{first.name}}")
{II}elif isinstance(first, IndexSegment):
{III}parts.append(f"[{{first.index}}]")
{II}else:
{III}aas_common.assert_never(first)

{II}for segment in iterator:
{III}if isinstance(segment, PropertySegment):
{IIII}parts.append(f".{{segment.name}}")
{III}elif isinstance(segment, IndexSegment):
{IIII}parts.append(f"[{{segment.index}}]")
{III}else:
{IIII}aas_common.assert_never(segment)

{II}return "".join(parts)"""
        ),
        Stripped(
            f"""\
class DeserializationException(Exception):
{I}\"\"\"Signal that the JSON de-serialization could not be performed.\"\"\"

{I}#: Human-readable explanation of the exception's cause
{I}cause: Final[str]

{I}#: Relative path to the erroneous value
{I}path: Final[Path]

{I}def __init__(
{III}self,
{III}cause: str
{I}) -> None:
{II}\"\"\"Initialize with the given :paramref:`cause` and an empty path.\"\"\"
{II}self.cause = cause
{II}self.path = Path()"""
        ),
        Stripped(
            f"""\
# NOTE (mristin, 2022-10-03):
# Recursive definitions are not yet available in mypy
# (see https://github.com/python/mypy/issues/731). We have to use ``Any``
# here, instead of recursive type annotations.
Jsonable = Union[
{I}bool,
{I}int,
{I}float,
{I}str,
{I}Sequence[Any],
{I}Mapping[str, Any]
]"""
        ),
        Stripped(
            f"""\
MutableJsonable = Union[
{I}bool,
{I}int,
{I}float,
{I}str,
{I}List[Any],
{I}MutableMapping[str, Any]
]"""
        ),
        Stripped("# region De-serialization"),
        _generate_bool_from_jsonable(),
        _generate_int_from_jsonable(),
        _generate_float_from_jsonable(),
        _generate_str_from_jsonable(),
        _generate_bytes_from_jsonable(),
        _generate_is_array_like(),
    ]  # type: List[Stripped]

    errors = []  # type: List[Error]

    for our_type in symbol_table.our_types:
        if isinstance(our_type, intermediate.Enumeration):
            blocks.append(_generate_enumeration_from_jsonable(enumeration=our_type))
        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            pass
        elif isinstance(our_type, intermediate.AbstractClass):
            blocks.append(_generate_dispatch_from_jsonable(cls=our_type))
        elif isinstance(our_type, intermediate.ConcreteClass):
            if len(our_type.concrete_descendants) > 0:
                blocks.append(_generate_dispatch_from_jsonable(cls=our_type))

            if our_type.is_implementation_specific:
                implementation_key = specific_implementations.ImplementationKey(
                    f"Jsonization/{our_type.name}_from_jsonable.py"
                )

                implementation = spec_impls.get(implementation_key, None)
                if implementation is None:
                    errors.append(
                        Error(
                            our_type.parsed.node,
                            f"The jsonization snippet is missing "
                            f"for the implementation-specific "
                            f"class {our_type.name}: {implementation_key}",
                        )
                    )
                    continue
            else:
                blocks.append(_generate_setter(cls=our_type))

                blocks.append(_generate_concrete_class_from_jsonable(cls=our_type))
        else:
            assert_never(our_type)

    # NOTE (mristin, 2022-10-03):
    # We add all the dispatch mappings at the end as the functions might not have been
    # defined yet otherwise.
    for our_type in symbol_table.our_types:
        if isinstance(our_type, intermediate.AbstractClass):
            blocks.append(_generate_dispatch_map_for_abstract_class(cls=our_type))
        elif isinstance(our_type, intermediate.ConcreteClass):
            if len(our_type.concrete_descendants) > 0:
                blocks.append(_generate_dispatch_map_for_concrete_class(cls=our_type))

            if not our_type.is_implementation_specific:
                blocks.append(_generate_setter_map(cls=our_type))

        else:
            pass

    blocks.append(Stripped("# endregion"))

    blocks.append(Stripped("# region Serialization"))

    blocks.append(_generate_bytes_to_base64_str())

    blocks.append(_generate_transformer(symbol_table=symbol_table))

    blocks.append(Stripped("_SERIALIZER = _Serializer()"))

    blocks.append(
        Stripped(
            f"""\
def to_jsonable(that: aas_types.Class) -> MutableJsonable:
{I}\"\"\"
{I}Convert :paramref:`that` to a JSON-able structure.

{I}:param that:
{II}AAS data to be recursively converted to a JSON-able structure
{I}:return:
{II}JSON-able structure which can be further encoded with, *e.g.*, :py:mod:`json`
{I}\"\"\"
{I}return _SERIALIZER.transform(that)"""
        )
    )

    blocks.append(Stripped("# endregion"))

    if len(errors) > 0:
        return None, errors

    blocks.append(python_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None



"""Generate C++ functions to verify an instance."""

from aas_core_codegen.cpp.verification import _generate

generate_header = _generate.generate_header
generate_implementation = _generate.generate_implementation



"""Provide a string representation of the regular expressions."""
from typing import Union, Optional

from aas_core_codegen import stringify
from aas_core_codegen.parse.retree import _types
from aas_core_codegen.parse.retree._types import (
    Char,
    Range,
    Concatenation,
    Symbol,
    Group,
    CharSet,
    Quantifier,
    Term,
    UnionExpr,
    Regex,
)
from aas_core_codegen.parse.tree import FormattedValue


def _stringify_char(that: Char) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("character", that.character),
            stringify.Property("explicitly_encoded", that.explicitly_encoded),
        ],
    )

    return result


def _stringify_range(that: Range) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("start", _stringify(that.start)),
            stringify.Property("end", _stringify(that.end)),
        ],
    )

    return result


def _stringify_concatenation(that: Concatenation) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("concatenants", list(map(_stringify, that.concatenants)))
        ],
    )

    return result


def _stringify_group(that: Group) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("union", _stringify(that.union)),
        ],
    )

    return result


def _stringify_char_set(that: CharSet) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("complementing", that.complementing),
            stringify.Property("ranges", list(map(_stringify, that.ranges))),
        ],
    )

    return result


def _stringify_quantifier(that: Quantifier) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("non_greedy", that.non_greedy),
            stringify.Property("minimum", that.minimum),
            stringify.Property("maximum", that.maximum),
        ],
    )

    return result


def _stringify_symbol(that: Symbol) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("kind", that.kind.value),
        ],
    )

    return result


def _stringify_term(that: Term) -> stringify.Entity:
    if isinstance(that.value, FormattedValue):
        stringified_value: Union[str, stringify.Entity, None] = "<formatted value>"
    else:
        stringified_value = _stringify(that.value)

    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("value", stringified_value),
            stringify.Property("quantifier", _stringify(that.quantifier)),
        ],
    )

    return result


def _stringify_union_expr(that: UnionExpr) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("uniates", list(map(_stringify, that.uniates))),
        ],
    )

    return result


def _stringify_regex(that: Regex) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("union", _stringify(that.union)),
        ],
    )

    return result


Dumpable = Union[
    Char,
    Range,
    Concatenation,
    Group,
    CharSet,
    Quantifier,
    Symbol,
    Term,
    UnionExpr,
    Regex,
]

stringify.assert_all_public_types_listed_as_dumpables(
    dumpable=Dumpable, types_module=_types
)

_DISPATCH = {
    Char: _stringify_char,
    Range: _stringify_range,
    Concatenation: _stringify_concatenation,
    Group: _stringify_group,
    CharSet: _stringify_char_set,
    Quantifier: _stringify_quantifier,
    Symbol: _stringify_symbol,
    Term: _stringify_term,
    UnionExpr: _stringify_union_expr,
    Regex: _stringify_regex,
}

stringify.assert_dispatch_exhaustive(dispatch=_DISPATCH, dumpable=Dumpable)


def _stringify(that: Optional[Dumpable]) -> Optional[stringify.Entity]:
    """Dispatch to the correct ``_stringify_*`` method."""
    if that is None:
        return None

    stringify_func = _DISPATCH.get(that.__class__, None)
    if stringify_func is None:
        raise AssertionError(
            f"No stringify function could be found for the class {that.__class__}"
        )

    stringified = stringify_func(that)  # type: ignore
    assert isinstance(stringified, stringify.Entity)
    stringify.assert_compares_against_dict(stringified, that)

    return stringified


def dump(that: Optional[Dumpable]) -> str:
    """Produce a string representation of the ``dumpable`` for testing or debugging."""
    if that is None:
        return repr(None)

    stringified = _stringify(that)
    return stringify.dump(stringified)



"""Provide a visitor implementation so that we do not have to duplicate the code."""
from aas_core_codegen.parse.retree._types import (
    Visitor,
    Regex,
    UnionExpr,
    Concatenation,
    Term,
    Group,
    Char,
    Quantifier,
    CharSet,
    Range,
    Symbol,
)
from aas_core_codegen.parse.tree import FormattedValue


class PassThroughVisitor(Visitor):
    """Visit all the nodes recursively without any action."""

    def visit_regex(self, node: Regex) -> None:
        """Visit the ``regex``."""
        self.visit(node.union)

    def visit_union_expr(self, node: UnionExpr) -> None:
        """Visit the ``union_expr``."""
        for concatenation in node.uniates:
            self.visit(concatenation)

    def visit_concatenation(self, node: Concatenation) -> None:
        """Visit the ``concatenation``."""
        for concatenant in node.concatenants:
            self.visit(concatenant)

    def visit_symbol(self, node: Symbol) -> None:
        """Visit the ``symbol``."""
        # NOTE (mristin, 2022-06-10):
        # The recursion stops here.

    def visit_term(self, node: Term) -> None:
        """Visit the ``term``."""
        if not isinstance(node.value, FormattedValue):
            self.visit(node.value)

        if node.quantifier is not None:
            self.visit(node.quantifier)

    def visit_group(self, node: Group) -> None:
        """Visit the ``group``."""
        self.visit(node.union)

    def visit_char(self, node: Char) -> None:
        """Visit the ``char``."""
        # NOTE (mristin, 2022-06-10):
        # The recursion stops here.

    def visit_quantifier(self, node: Quantifier) -> None:
        """Visit the ``quantifier``."""
        # NOTE (mristin, 2022-06-10):
        # The recursion stops here.

    def visit_char_set(self, node: CharSet) -> None:
        """Visit the ``char_set``."""
        for a_range in node.ranges:
            self.visit(a_range)

    def visit_range(self, node: Range) -> None:
        """Visit the ``range``."""
        # NOTE (mristin, 2022-06-10):
        # The recursion stops here.



"""Understand the constructors of the classes."""
import abc
import ast
import collections
import itertools
import pathlib
from typing import (
    List,
    Optional,
    MutableMapping,
    Mapping,
    Sequence,
    Union,
    Tuple,
    AbstractSet,
)

import asttokens
from icontract import ensure, require, DBC

from aas_core_codegen import parse, stringify
from aas_core_codegen.common import (
    Identifier,
    Error,
    assert_union_of_descendants_exhaustive,
)
from aas_core_codegen.parse import Method

_MODULE_NAME = pathlib.Path(__file__).parent.name


class CallSuperConstructor:
    """
    Represent a call to the constructor of a super class.

    The arguments of the original ``__init__`` are expected to be propagated as-are.
    """

    super_name: Identifier  #: Identifier of the super class

    def __init__(self, super_name: Identifier) -> None:
        """Initialize with the given values."""
        self.super_name = super_name


class Default(DBC):
    """Represent a default value set to a property if an argument is unspecified."""

    def __init__(self, node: ast.AST) -> None:
        """Initialize with the given values."""
        self.node = node

    @abc.abstractmethod
    def __repr__(self) -> str:
        # Signal that the class is purely abstract
        raise NotImplementedError()


class EmptyList(Default):
    """Represent an empty list set to a property if the argument is unspecified."""

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return f"<{_MODULE_NAME}.{self.__class__.__name__} at 0x{id(self):x}>"


class DefaultEnumLiteral(Default):
    """Represent an enum literal set to a property if the argument is unspecified."""

    @require(lambda enum, literal: literal in enum.literals)
    def __init__(
        self, enum: parse.Enumeration, literal: parse.EnumerationLiteral, node: ast.AST
    ) -> None:
        """Initialize with the given values."""
        Default.__init__(self, node=node)
        self.enum = enum
        self.literal = literal

    def __repr__(self) -> str:
        """Represent the instance as a string for easier debugging."""
        return f"<{_MODULE_NAME}.{self.__class__.__name__} at 0x{id(self):x}>"


class AssignArgument:
    """Represent an assignment of an argument to ``__init__`` to a property."""

    name: Identifier  #: Identifier of the property
    argument: Identifier  #: Identifier of the argument
    default: Optional["DefaultUnion"]  #: Default value if the argument is None

    def __init__(
        self, name: Identifier, argument: Identifier, default: Optional["DefaultUnion"]
    ) -> None:
        """Initialize with the given values."""
        self.name = name
        self.argument = argument
        self.default = default


Statement = Union[CallSuperConstructor, AssignArgument]


# noinspection PyTypeChecker
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _call_as_call_to_super_init(
    call: ast.Call,
    parsed_class: parse.Class,
    parsed_symbol_table: parse.SymbolTable,
    atok: asttokens.ASTTokens,
) -> Tuple[Optional[CallSuperConstructor], Optional[Error]]:
    """Understand a call as a call to the constructor of a super-class."""
    if not isinstance(call.func, ast.Attribute):
        return (
            None,
            Error(
                call,
                f"Unexpected call in the body "
                f"of ``__init__``: {atok.get_text(call.func)}; "
                f"only calls to super ``__init__``'s are expected",
            ),
        )

    if call.func.attr != "__init__":
        return (
            None,
            Error(
                call,
                f"Unexpected call in the body "
                f"of ``__init__``: {atok.get_text(call.func)}; "
                f"only calls to super ``__init__``'s are expected",
            ),
        )

    if not isinstance(call.func.value, ast.Name):
        return (
            None,
            Error(
                call.func.value,
                f"Expected a super class as a name "
                f"for a call to super ``__init__``, "
                f"but got: {atok.get_text(call.func.value)}",
            ),
        )

    identifier = Identifier(call.func.value.id)

    if identifier not in parsed_class.inheritances:
        return (
            None,
            Error(
                call.func.value,
                f"Expected a super class in the call "
                f"to a super ``__init__``, "
                f"but {parsed_class.name} does not inherit "
                f"from {identifier}",
            ),
        )

    parsed_super_class = parsed_symbol_table.must_find_class(name=identifier)

    if "__init__" not in parsed_super_class.methods_by_name:
        return (
            None,
            Error(
                call.func,
                f"The super class {parsed_super_class.name} "
                f"does not define a ``__init__``",
            ),
        )

    # region Check the arguments of the call to super ``__init__``

    double_star_keyword = next(
        (keyword for keyword in call.keywords if keyword.arg is None), None
    )

    if double_star_keyword is not None:
        return (
            None,
            Error(
                double_star_keyword,
                "Expected a call to a super ``__init__`` to provide only "
                "explicit keyword arguments, "
                "but got a double-star keyword argument",
            ),
        )

    underlying_errors = []  # type: List[Error]

    for arg_node in itertools.chain(
        call.args, (keyword.value for keyword in call.keywords)
    ):
        if not isinstance(arg_node, ast.Name):
            underlying_errors.append(
                Error(
                    arg_node,
                    f"Expected only names in the arguments to "
                    f"super ``__init__``, but got: {atok.get_text(arg_node)}",
                )
            )

    if len(underlying_errors) > 0:
        return (
            None,
            Error(
                call,
                "Failed to parse the arguments to the super ``__init__``",
                underlying_errors,
            ),
        )

    super_init = parsed_super_class.methods_by_name[Identifier("__init__")]
    resolved_kwargs = dict()  # type: MutableMapping[str, str]

    if len(call.args) > len(super_init.arguments):
        return (
            None,
            Error(
                call,
                f"The ``{parsed_super_class.name}.__init__`` "
                f"expected {len(super_init.arguments)} argument(s), "
                f"but the call provides "
                f"{len(call.args)} positional argument(s)",
            ),
        )

    for arg_node, argument in zip(call.args, super_init.arguments):
        assert isinstance(arg_node, ast.Name)
        resolved_kwargs[argument.name] = arg_node.id

    assert len(underlying_errors) == 0

    for keyword in call.keywords:
        if keyword.arg not in super_init.arguments_by_name:
            underlying_errors.append(
                Error(
                    keyword,
                    f"The ``{parsed_super_class.name}.__init__`` does not expect "
                    f"the argument {keyword.arg}",
                )
            )
        else:
            assert isinstance(keyword.value, ast.Name)
            assert isinstance(keyword.arg, str)
            resolved_kwargs[keyword.arg] = keyword.value.id

    if len(underlying_errors) > 0:
        return (
            None,
            Error(
                call,
                "Failed to parse the arguments to the super ``__init__``",
                underlying_errors,
            ),
        )

    init = parsed_class.methods_by_name[Identifier("__init__")]
    for key, val in resolved_kwargs.items():
        if val not in init.arguments_by_name:
            underlying_errors.append(
                Error(
                    call,
                    f"Expected all the arguments to "
                    f"``{parsed_super_class.name}.__init__`` "
                    f"to be propagation of the original ``__init__`` "
                    f"arguments, but the name {key} is not an argument "
                    f"of ``{parsed_class.name}.__init__``",
                )
            )

        elif key != val:
            underlying_errors.append(
                Error(
                    call,
                    f"Expected the arguments to super ``__init__`` "
                    f"to be passed with the same names, "
                    f"but the argument {key} is passed "
                    f"as the name {val}",
                )
            )

    missing_args = [
        argument.name
        for argument in super_init.arguments
        if argument.name not in resolved_kwargs
    ]

    if len(missing_args) > 0:
        missing_args_str = ", ".join(missing_args)
        underlying_errors.append(
            Error(
                call,
                f"The call to ``{parsed_super_class.name}.__init__`` "
                f"is missing one or more arguments: "
                f"{missing_args_str}",
            )
        )

    if len(underlying_errors) > 0:
        return (
            None,
            Error(
                call,
                "Failed to parse the arguments to the super ``__init__``",
                underlying_errors,
            ),
        )
    # endregion

    return CallSuperConstructor(super_name=parsed_super_class.name), None


# noinspection PyTypeChecker
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _understand_assignment(
    assign: ast.Assign,
    init: Method,
    parsed_class: parse.Class,
    parsed_symbol_table: parse.SymbolTable,
    atok: asttokens.ASTTokens,
) -> Tuple[Optional[Statement], Optional[Error]]:
    if len(assign.targets) > 1:
        return (
            None,
            Error(
                assign,
                f"Expected only a single target for property assignment, "
                f"but got {len(assign.targets)} targets",
            ),
        )

    target = assign.targets[0]

    if not (
        isinstance(target, ast.Attribute)
        and isinstance(target.value, ast.Name)
        and target.value.id == "self"
    ):
        return (
            None,
            Error(
                target,
                f"Expected a property as the target of an assignment, "
                f"but got: {atok.get_text(target)}",
            ),
        )

    if target.attr not in parsed_class.properties_by_name:
        return (
            None,
            Error(
                target.value,
                f"The property has not been previously "
                f"defined in the class {parsed_class.name!r}: {target.attr}",
            ),
        )

    if isinstance(assign.value, ast.Name):
        if assign.value.id not in init.arguments_by_name:
            return (
                None,
                Error(
                    assign.value,
                    f"Expected the property {target.attr} to be assigned "
                    f"to an argument, but it was assigned to a non-argument variable: "
                    f"{atok.get_text(assign.value)}",
                ),
            )

        if target.attr != assign.value.id:
            return (
                None,
                Error(
                    assign.value,
                    f"Expected the property {target.attr} to be assigned "
                    f"exactly the argument with the same name, "
                    f"but got: {atok.get_text(assign.value)}",
                ),
            )

        return (
            AssignArgument(
                name=Identifier(target.attr),
                argument=Identifier(target.attr),
                default=None,
            ),
            None,
        )
    elif isinstance(assign.value, ast.IfExp):
        default_node = None  # type: Optional[ast.AST]

        if_exp = assign.value
        # noinspection PyUnresolvedReferences
        if (
            isinstance(if_exp.test, ast.Compare)
            and isinstance(if_exp.test.left, ast.Name)
            and isinstance(if_exp.test.left.ctx, ast.Load)
            and if_exp.test.left.id in init.arguments_by_name
            and len(if_exp.test.ops) == 1
            and isinstance(if_exp.test.ops[0], ast.IsNot)
            and len(if_exp.test.comparators) == 1
            and isinstance(if_exp.test.comparators[0], ast.Constant)
            and if_exp.test.comparators[0].value is None
            and isinstance(if_exp.body, ast.Name)
            and if_exp.body.id == if_exp.test.left.id
            and if_exp.orelse is not None
        ):
            default_node = if_exp.orelse

        elif (
            isinstance(if_exp.test, ast.Compare)
            and isinstance(if_exp.test.left, ast.Name)
            and if_exp.test.left.id in init.arguments_by_name
            and len(if_exp.test.ops) == 1
            and isinstance(if_exp.test.ops[0], ast.Is)
            and len(if_exp.test.comparators) == 1
            and isinstance(if_exp.test.comparators[0], ast.Constant)
            and if_exp.test.comparators[0].value is None
            and isinstance(if_exp.orelse, ast.Name)
            and if_exp.orelse.id == if_exp.test.left.id
            and if_exp.body is not None
        ):
            default_node = if_exp.body
        else:
            # We did not match this case.
            pass

        if default_node is not None:
            default = None  # type: Optional[DefaultUnion]

            if isinstance(default_node, ast.List) and default_node.elts == []:
                default = EmptyList(node=default_node)

            elif isinstance(default_node, ast.Attribute) and isinstance(
                default_node.value, ast.Name
            ):
                our_type = parsed_symbol_table.find_our_type(
                    name=Identifier(default_node.value.id)
                )

                if isinstance(our_type, parse.Enumeration):
                    literal = our_type.literals_by_name.get(
                        Identifier(default_node.attr), None
                    )

                    if literal is not None:
                        default = DefaultEnumLiteral(
                            enum=our_type, literal=literal, node=default_node
                        )
            else:
                assert default is None

            if default is None:
                return None, Error(
                    if_exp.orelse,
                    f"The handling of this default value for "
                    f"the property {target.attr!r} has not been implemented: "
                    f"{ast.dump(default_node)}",
                )
            else:
                assert isinstance(if_exp.test, ast.Compare)
                assert isinstance(if_exp.test.left, ast.Name)

                return (
                    AssignArgument(
                        name=Identifier(target.attr),
                        argument=Identifier(if_exp.test.left.id),
                        default=default,
                    ),
                    None,
                )

    return None, Error(
        assign,
        f"The handling of the constructor statement "
        f"has not been implemented: {ast.dump(assign)}; "
        f"please notify the developers if you really need this feature",
    )


# noinspection PyTypeChecker
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _understand_body(
    parsed_class: parse.Class,
    parsed_symbol_table: parse.SymbolTable,
    atok: asttokens.ASTTokens,
) -> Tuple[Optional[List[Statement]], Optional[Error]]:
    """Try to understand the body of the constructor for the given ``parsed_class``."""
    init = None  # type: Optional[parse.Method]
    for method in parsed_class.methods:
        if method.name == "__init__":
            init = method
            break

    if init is None:
        return [], None

    errors = []  # type: List[Error]
    result = []  # type: List[Statement]

    assert isinstance(init.node, ast.FunctionDef)

    for stmt in init.node.body:
        if isinstance(stmt, ast.Pass):
            continue
        elif isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Call):
            call_super_init, error = _call_as_call_to_super_init(
                call=stmt.value,
                parsed_class=parsed_class,
                parsed_symbol_table=parsed_symbol_table,
                atok=atok,
            )

            if error is not None:
                errors.append(error)
            else:
                assert call_super_init is not None
                result.append(call_super_init)

        elif isinstance(stmt, ast.Assign):
            prop_assignment, error = _understand_assignment(
                assign=stmt,
                init=init,
                parsed_class=parsed_class,
                parsed_symbol_table=parsed_symbol_table,
                atok=atok,
            )

            if error is not None:
                errors.append(error)
            else:
                assert prop_assignment is not None
                result.append(prop_assignment)
        else:
            errors.append(
                Error(
                    stmt,
                    f"Unexpected statement in the body "
                    f"of ``__init__``: {atok.get_text(stmt)}; "
                    f"only calls to super ``__init__``'s and "
                    f"property assignments expected",
                )
            )

    if len(errors) > 0:
        return (
            None,
            Error(
                init.node,
                f"Failed to understand the constructor "
                f"of the class {parsed_class.name}",
                underlying=errors,
            ),
        )

    return result, None


class ConstructorTable:
    """Map understanding of constructors for the classes."""

    def __init__(self, mapping: Mapping[parse.Class, Sequence[Statement]]) -> None:
        self._mapping = mapping

    def has(self, parsed_class: parse.Class) -> bool:
        """Check if there is an entry in the table for the given ``parsed_class``."""
        return parsed_class in self._mapping

    def must_find(self, parsed_class: parse.Class) -> Sequence[Statement]:
        """
        Find the constructor corresponding to this ``parsed_class``.

        :raise: :py:attr:`KeyError` if the entry does not exist.
        """
        result = self._mapping.get(parsed_class, None)
        if result is None:
            raise KeyError(
                f"No entry found in the constructor table for the class: {parsed_class}"
            )

        return result

    def entries(self) -> AbstractSet[Tuple[parse.Class, Sequence[Statement]]]:
        """Retrieve all the entries in the table."""
        # noinspection PyTypeChecker
        return self._mapping.items()


# fmt: off
@ensure(
    lambda parsed_symbol_table, result:
    result[0] is None
    or all(
        result[0].has(our_type)
        for our_type in parsed_symbol_table.our_types
        if isinstance(our_type, parse.Class)
    ),
    "Constructor understood for each class"
)
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
# fmt: on
def understand_all(
    parsed_symbol_table: parse.SymbolTable, atok: asttokens.ASTTokens
) -> Tuple[Optional[ConstructorTable], Optional[Error]]:
    """Understand the constructors of all the classes in the symbol table."""
    errors = []  # type: List[Error]
    mapping = (
        collections.OrderedDict()
    )  # type: MutableMapping[parse.Class, List[Statement]]

    for our_type in parsed_symbol_table.our_types:
        if not isinstance(our_type, parse.Class):
            continue

        statements, error = _understand_body(
            parsed_class=our_type, parsed_symbol_table=parsed_symbol_table, atok=atok
        )

        if error is not None:
            errors.append(error)
        else:
            assert statements is not None
            mapping[our_type] = statements

    if len(errors) > 0:
        return (
            None,
            Error(
                atok.tree, "Failed to understand the constructors", underlying=errors
            ),
        )

    return ConstructorTable(mapping=mapping), None


# NOTE (mristin, 2021-12-26):
# At this point we need to dump only a subset of this module, so we go for YAGNI
# principle and do not provide dump functions for everything.

Dumpable = Union[
    Default,
    EmptyList,
    DefaultEnumLiteral,
    AssignArgument,
    CallSuperConstructor,
]


def _stringify_empty_list(
    that: EmptyList,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.PropertyEllipsis("node", that.node),
        ],
    )

    return result


def _stringify_default_enum_literal(
    that: DefaultEnumLiteral,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.PropertyEllipsis("node", that.node),
            stringify.Property(
                "enum",
                f"Reference to {that.enum.__class__.__name__} {that.enum.name}",
            ),
            stringify.Property(
                "literal",
                f"Reference to {that.literal.__class__.__name__} "
                f"{that.literal.name}",
            ),
        ],
    )

    return result


def _stringify_assign_argument(
    that: AssignArgument,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property("argument", that.argument),
            stringify.Property("default", _stringify(that.default)),
        ],
    )

    return result


def _stringify_call_super_constructor(
    that: CallSuperConstructor,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("super_name", that.super_name),
        ],
    )

    return result


_DISPATCH = {
    EmptyList: _stringify_empty_list,
    DefaultEnumLiteral: _stringify_default_enum_literal,
    AssignArgument: _stringify_assign_argument,
    CallSuperConstructor: _stringify_call_super_constructor,
}

stringify.assert_dispatch_exhaustive(dispatch=_DISPATCH, dumpable=Dumpable)


def _stringify(that: Optional[Dumpable]) -> Optional[stringify.Entity]:
    """Dispatch to the correct ``_stringify_*`` method."""
    if that is None:
        return None

    stringify_func = _DISPATCH.get(that.__class__, None)
    if stringify_func is None:
        raise AssertionError(
            f"No stringify function could be found for the class {that.__class__}"
        )

    stringified = stringify_func(that)  # type: ignore
    assert isinstance(stringified, stringify.Entity)
    stringify.assert_compares_against_dict(stringified, that)

    return stringified


def dump(that: Optional[Dumpable]) -> str:
    """Produce a string representation of the ``dumpable`` for testing or debugging."""
    if that is None:
        return repr(None)

    stringified = _stringify(that)
    return stringify.dump(stringified)


DefaultUnion = Union[EmptyList, DefaultEnumLiteral]
assert_union_of_descendants_exhaustive(union=DefaultUnion, base_class=Default)
